from django.contrib.auth.models import User
from django.db.models import Q, Count
from django.shortcuts import get_object_or_404

from inventory.models import InventoryObject, Lab
from racks import crud
from django.views.generic.base import ContextMixin


class GetAccountMixin(ContextMixin):
    def get_object(self, **kwargs):
        return get_object_or_404(User, username=kwargs['username'])


class AccountsListView(crud.ListView):
    def get_queryset(self):
        return self.model.objects.all().annotate(
            inventory_objects_count=Count('assigned_objects'),
            labs_count=Count('assigned_labs'),
        ).order_by('-inventory_objects_count', 'username')


class AccountDetailView(GetAccountMixin, crud.DetailView):
    template = 'profile'


class AccountAssignedDetailView(GetAccountMixin, crud.DetailView):
    template = 'assigned'

    def get_context_data(self, **kwargs):
        context = super(AccountAssignedDetailView, self).get_context_data(**kwargs)
        user = self.get_object(**kwargs)
        user_groups = user.groups.all()

        context.update({
            'assigned_objects':
                InventoryObject.objects.filter(
                    Q(assigned_user=user) | Q(assigned_group__in=user_groups)
                ).distinct(),
            'assigned_labs':
                Lab.objects.filter(
                    Q(assigned_user=user) | Q(assigned_group__in=user_groups)
                ).distinct(),
        })
        return context


class AccountResponsibleDetailView(GetAccountMixin, crud.DetailView):
    template = 'responsible'

    def get_context_data(self, **kwargs):
        context = super(AccountResponsibleDetailView, self).get_context_data(**kwargs)
        user = self.get_object(**kwargs)
        context.update({
            'inventory_objects': user.inventory_objects.all() if user else None,
            'labs': user.labs.all() if user else None,
        })
        return context


class AccountViewSet(crud.ViewSet):
    PATTERNS = (
        (r'^$', 'list'),
        (r'^(?P<username>[0-9A-Za-z_@\-\+\.]{1,30})/$', 'profile'),
        (r'^(?P<username>[0-9A-Za-z_@\-\+\.]{1,30})/assigned$', 'assigned'),
        (r'^(?P<username>[0-9A-Za-z_@\-\+\.]{1,30})/responsible$', 'responsible'),
    )

    model = User
    list_view = AccountsListView
    profile_view = AccountDetailView
    assigned_view = AccountAssignedDetailView
    responsible_view = AccountResponsibleDetailView
    fields = '__all__'


urlpatterns = AccountViewSet('account').urlpatterns
