from django.conf import settings
from django.db import models
from django.utils.functional import cached_property
from django.utils.http import urlencode

from racks.models import RacksModel

DCIM_EMPTY = 0
DCIM_SERVER = 1
DCIM_SWITCH = 4
DCIM_OTHER = -1

# Only DCIM_SERVER and DCIM_SWITCH are currently imported. All others are mapped to DCIM_OTHER.
DCIM_DEVICE_TYPES = (
    (DCIM_EMPTY, 'Empty'),
    (DCIM_SERVER, 'Server'),
    (2, 'Appliance'),
    (3, 'Storage Array'),
    (DCIM_SWITCH, 'Switch'),
    (5, 'Chassis'),
    (6, 'Patch Panel'),
    (7, 'Physical Infrastructure'),
    (DCIM_OTHER, 'Other Object'),
)


class Switch(RacksModel):
    dcim_id = models.IntegerField(primary_key=True)
    label = models.CharField(max_length=255)
    management_address = models.CharField(max_length=255)

    def __unicode__(self):
        return self.label

    def active_ports(self):
        return [p for p in self.ports.all() if p.dcim_device_type != DCIM_EMPTY]

    def empty_ports(self):
        return [p for p in self.ports.all() if p.dcim_device_type == DCIM_EMPTY]

    def ports_to_servers(self):
        return [p for p in self.ports.all() if p.dcim_device_type == DCIM_SERVER]

    def ports_to_switches(self):
        return [p for p in self.ports.all() if p.dcim_device_type == DCIM_SWITCH]

    def ports_to_others(self):
        return [p for p in self.ports.all() if p.dcim_device_type == DCIM_OTHER]

    @cached_property
    def locations(self):
        locations = set()
        for port in self.ports.all():
            for address in port.l2addresses.all():
                if address.inventory_object and address.inventory_object.location_name:
                    locations.add(address.inventory_object.location_name)
        return sorted(locations)

    @property
    def dcim_link(self):
        return settings.DCIM_URL + 'devices.php?' + urlencode({'DeviceID': self.dcim_id})

    @property
    def ssh_link(self):
        if self.management_address:
            return 'ssh://%s' % self.management_address
        return None

    PUBLIC_FIELDS = (
        'dcim_id',
        'label',
        'ports',
    )


class SwitchPort(RacksModel):
    number = models.CharField(max_length=255, null=True)
    label = models.CharField(max_length=255, null=True)
    switch = models.ForeignKey(Switch, related_name='ports')
    l2addresses = models.ManyToManyField('inventory.L2Address', related_name='switch_ports')
    dcim_server_id = models.IntegerField(null=True)
    switch_port_name = models.CharField(max_length=255, null=True)
    dcim_device_type = models.IntegerField(default=DCIM_EMPTY, choices=DCIM_DEVICE_TYPES)

    def l2addresses_with_objects(self):
        return [a for a in self.l2addresses.all() if a.inventory_object]

    def l2addresses_without_objects(self):
        return [a for a in self.l2addresses.all() if not a.inventory_object]

    class Meta:
        unique_together = ('switch', 'number')

    def __unicode__(self):
        if self.label is not None:
            return u'Port "{}"'.format(self.label)
        if self.switch_port_name is not None:
            return u'Port "{}"'.format(self.switch_port_name)
        if self.number is not None:
            return u'Port #{}'.format(self.number)
        return u'Unlabeled Port (UID #{}) '.format(self.pk)

    PUBLIC_FIELDS = (
        'id',
        'number',
        'label',
        'switch',
        'l2addresses',
        'dcim_server_id',
        'switch_port_name',
    )
