import logging
import os
import requests
import urllib
import urlparse
import yaml

from datetime import datetime
from pytz import UTC
from xml.etree import ElementTree

from django.conf import settings
from django.core.exceptions import ValidationError, ImproperlyConfigured
from django.core.validators import validate_ipv46_address
from django.db import connection

from jenkins_ci.models import JenkinsInstance, JenkinsNode, JenkinsLabel


def utcnow():
    return datetime.utcnow().replace(tzinfo=UTC)


class JenkinsImportError(Exception):

    status_code = None

    def __init__(self, code=None):
        self.status_code = code
        super(JenkinsImportError, self).__init__(self)


def _create_log_wrapper(severity, fail_by_default=False):
    def wrapped_method(self, message, *args, **kwargs):
        getattr(self.logger, severity)(message, *args)
        instance = kwargs.get('instance')
        if instance:
            instance.error_logs.create(
                instance_name=instance.name,
                instance_url=instance.url,
                message=message % args,
            )
            if kwargs.get('fail', fail_by_default):
                instance.last_failed_sync = utcnow()
                instance.save()
    return wrapped_method


class JenkinsLogger(object):

    def __init__(self, name):
        self.logger = logging.getLogger(name)

    exception = _create_log_wrapper('exception', fail_by_default=True)
    critical = _create_log_wrapper('critical', fail_by_default=True)
    error = _create_log_wrapper('error')
    warning = _create_log_wrapper('warning')
    info = _create_log_wrapper('info')
    debug = _create_log_wrapper('debug')


log = JenkinsLogger('importers.jenkins')


def load_credentials():
    try:
        conf_dirs = settings.SITE_CONF_DIRS
    except AttributeError:
        raise ImproperlyConfigured('Credentials parameters not found in settings file.')

    for conf_dir in conf_dirs:
        full_path = os.path.join(conf_dir, 'jenkins.yaml')
        if os.path.exists(full_path):
            with open(full_path) as f:
                return yaml.safe_load(f)

    raise ImproperlyConfigured("Credentials file 'jenkins.yaml' not found in %r" % (conf_dirs))


def jenkins_api(instance, username=None, api_key=None):
    if username is None:
        username = instance.username
    if api_key is None:
        api_key = instance.api_key

    def jenkins_api_request(query, format='text'):
        url = ''.join((instance.url.rstrip('/'), '/', query))
        log.debug('Requesting Jenkins API at %s', url)
        try:
            response = requests.get(url, auth=(username, api_key))
        except requests.exceptions.RequestException:
            log.exception('Failed to retrieve Jenkins API at %s', url, instance=instance)
            raise JenkinsImportError()
        if response.status_code != requests.codes.ok:
            log.critical(
                'Jenkins API response {%d} for %s',
                response.status_code,
                url,
                instance=instance
            )
            raise JenkinsImportError(response.status_code)
        if format == 'json':
            try:
                return response.json()
            except ValueError:
                log.exception('Failed to unparse Jenkins API response: not JSON', instance=instance)
                raise JenkinsImportError()
        elif format == 'xml':
            try:
                return ElementTree.fromstring(response.text)
            except Exception:
                log.exception('Failed to unparse Jenkins API response: not XML', instance=instance)
                raise JenkinsImportError()
        else:
            return response.text

    return jenkins_api_request


def import_jenkins(instance, api):

    log.info('Importing Jenkins CI %r...', instance.name)

    # Retrieve Jenkins nodes
    try:
        nodes = api('computer/api/json', format='json')['computer']
    except JenkinsImportError:
        return
    except KeyError:
        log.exception('Jenkins API response does not contain slaves list', instance=instance)
        return

    # Retrieve Jenkins last job builds and build node->timestamp dictionary to update node last jobs
    node_last_jobs = {}
    try:
        last_jobs = api('api/json?tree=jobs[lastBuild[timestamp,builtOn]]', format='json')['jobs']
    except JenkinsImportError:
        return
    except KeyError:
        log.error('Jenkins API response does not contain jobs list', instance=instance)
    else:
        for job in last_jobs:
            if not job.get('lastBuild'):
                continue
            node_name = job['lastBuild']['builtOn']
            timestamp = job['lastBuild']['timestamp']
            node_last_jobs[node_name] = max(node_last_jobs.get(node_name, 0), timestamp)

    # Process all nodes individually
    config_xml_accessible = True
    for computer in nodes:
        log.debug('Processing Jenkins Node `%s`', computer['displayName'])

        # STEP 1. Parse values from config.xml file if available

        # Initialize default values
        node_name = computer['displayName']
        node_labels = []
        node_host = None
        node_ip = None
        node_master = False

        # Update host/IP/labels from from config.xml if it's available
        if config_xml_accessible:

            if node_name == u'master':
                url = 'computer/(master)/config.xml'
            else:
                url = '/'.join(('computer', urllib.quote(node_name), 'config.xml'))

            try:
                response = api(url, format='xml')
            except JenkinsImportError as e:
                # Error is already logged in jenkins_api_request, so we only need to do some extra
                # processing for HTTP response code 403.
                if e.status_code == 403:
                    log.error(
                        'Error 403 Forbidden: check Administrate privileges on Jenkins.',
                        instance=instance,
                    )
                    # We won't try retrieving config.xml from other nodes on the same instance
                    config_xml_accessible = False
            else:
                # XML parser may insert namespaces into tag names, so we have to use `.endswith`
                for element in response:
                    if element.tag.endswith('label'):
                        if not element.text:
                            continue
                        node_labels = filter(
                            None,
                            map(
                                lambda st: st.strip(),
                                element.text.split(' ')
                            )
                        )
                    elif element.tag.endswith('launcher'):
                        for subelement in element:
                            if subelement.tag.endswith('host'):
                                try:
                                    validate_ipv46_address(subelement.text)
                                    node_ip = subelement.text
                                except Exception:
                                    node_host = subelement.text
                                break

        # Detect if it's a master node. If true, update node host/IP from instance.url.
        if node_name == u'master':
            node_master = True
            jenkins_addr = urlparse.urlparse(instance.url).hostname
            try:
                validate_ipv46_address(jenkins_addr)
            except ValidationError:
                node_host = jenkins_addr
            else:
                node_ip = jenkins_addr

        # Try to match to an existing node record if possible, otherwise create new one.
        node, created = instance.nodes.get_or_create(name=node_name)

        # Apply field values from config.xml
        node.hostname = node_host
        node.address = node_ip

        # Update node's last_job timestamp if there's a matching value in node_last_jobs dict.
        last_job = node_last_jobs.get(node_name)
        if last_job:
            last_job = datetime.utcfromtimestamp(last_job / 1000).replace(tzinfo=UTC)
            node.last_job = max(node.last_job, last_job) if node.last_job else last_job

        # Apply labels from config.xml, trying to re-use already existing ones
        had_labels = set(l.name for l in node.labels.all())
        del_labels = had_labels - set(node_labels)
        add_labels = set(node_labels) - had_labels

        node.labels.remove(*node.labels.filter(name__in=del_labels))
        for label_name in add_labels:
            label, created = JenkinsLabel.objects.get_or_create(name=label_name)
            node.labels.add(label)

        # STEP 2. Update node record with data from API response.

        monitors = computer.get('monitorData') or {}  # Convert None to {}

        node.offline = computer.get('offline', True)
        node.temporarily_offline = computer.get('temporarilyOffline', False)
        node.offline_reason = computer.get('offlineCauseReason') or None  # Convert "" to None
        node.executors = computer.get('numExecutors', 0)

        node.architecture = monitors.get('hudson.node_monitors.ArchitectureMonitor')

        # Check monitor data. Monitor fields may be None in JSON, so `or {}` syntax is used
        # instead of providing a default `{}` value to `get()` method.

        monitor = monitors.get('hudson.node_monitors.SwapSpaceMonitor') or {}
        node.available_memory = monitor.get('availablePhysicalMemory', 0)
        node.total_memory = monitor.get('totalPhysicalMemory', 0)
        node.available_swap = monitor.get('availableSwapSpace', 0)
        node.total_swap = monitor.get('totalSwapSpace', 0)

        monitor = monitors.get('hudson.node_monitors.DiskSpaceMonitor') or {}
        node.disk_space = monitor.get('size', 0)

        monitor = monitors.get('hudson.node_monitors.TemporarySpaceMonitor') or {}
        node.temp_disk_space = monitor.get('size', 0)

        # Mark node as active and save

        node.disabled = False
        node.save()

        # Oh, and if this was a master node, update instance record.

        if node_master:
            instance.master_node = node

    if config_xml_accessible:
        instance.last_successful_sync = utcnow()
    else:
        # If we failed to load node labels, it's a failed sync even if everything else worked fine.
        instance.last_failed_sync = utcnow()
    instance.save()


def import_all_jenkins():
    # Load/update any new Jenkins instances from credentials file.
    ci_to_import = []
    exclude_pk = []

    try:
        credentials = load_credentials()
    except ImproperlyConfigured as e:
        log.error('Failed to load Jenkins credentials, restricting import to DB only: %s', e)
    else:
        for record in credentials:
            ci, created = JenkinsInstance.objects.get_or_create(url=record['url'])
            if created or ci.name != record['name']:
                ci.name = record['name']
                ci.save()
            exclude_pk.append(ci.pk)
            ci_to_import.append((ci, jenkins_api(ci, record['username'], record['api_key'])))

    database_cis = JenkinsInstance.objects.filter(
        username__isnull=False,
        api_key__isnull=False,
    ).exclude(
        pk__in=exclude_pk,
    )
    ci_to_import.extend((ci, jenkins_api(ci)) for ci in database_cis)

    # Mark all nodes as disabled
    JenkinsNode.objects.update(disabled=True)

    # Import data from all jenkins
    for ci, api in ci_to_import:
        try:
            import_jenkins(ci, api)
        except Exception:
            log.exception('Unhandled exception during data import', instance=ci)
            ci.nodes.update(disabled=False)

    # Delete all unprocessed Nodes (i.e. those that no longer belong to any instance)
    JenkinsNode.objects.filter(disabled=True).delete()


def bind_jenkins_nodes_to_inventory():
    """
    Search for matches between JenkinsNode.hostname == InventoryObject.name and bind them together.

    We're using SQL here since the fields are not in ForeignKey relationship.

    :return: None
    """
    cursor = connection.cursor()
    cursor.execute(
        '''
        SELECT t1.id, t2.id
        FROM jenkins_ci_jenkinsnode t1
        INNER JOIN inventory_inventoryobject t2 ON t1.hostname = t2.name
        WHERE t1.inventory_object_id IS NULL AND t1.hostname IS NOT NULL
        '''
    )
    for jenkins_id, inventory_id in cursor.fetchall():
        node = JenkinsNode.objects.get(pk=jenkins_id)
        node.inventory_object_id = inventory_id
        node.save()

    unmatched_nodes = JenkinsNode.objects\
        .filter(inventory_object=None, hostname__isnull=False)\
        .select_related('instance')
    for node in unmatched_nodes:
        log.error(
            "Jenkins slave '%s' (hostname '%s'): no Inventory Object found with matching name",
            node.name,
            node.hostname,
            instance=node.instance,
        )
