import logging
from django.conf import settings
from django_auth_ldap.config import _LDAPConfig

from django.contrib.auth.models import User, Group
from racks.models import Source, SOURCE_LDAP

log = logging.getLogger(__name__)


def ldap_connection():
    ldap = _LDAPConfig.get_ldap()
    return ldap.initialize(settings.AUTH_LDAP_SERVER_URI)


def map_user_attrs(attrs):
    model_attrs = {
        'username': attrs[settings.LDAP_USERNAME_ATTR][0],
    }

    for model_attr, ldap_attr in settings.AUTH_LDAP_USER_ATTR_MAP.items():
        if ldap_attr in attrs:
            model_attrs[model_attr] = attrs[ldap_attr][0]

    return model_attrs


def get_ldap_users():
    connection = ldap_connection()
    user_search = settings.LDAP_USER_SEARCH

    result = user_search.execute(connection)

    return [
        (dn, map_user_attrs(attrs))
        for dn, attrs in result
    ]


def get_ldap_groups():
    connection = ldap_connection()
    group_search = settings.AUTH_LDAP_GROUP_SEARCH

    result = group_search.execute(connection)

    return [attrs for dn, attrs in result]


def import_users():

    users = {
        u.username: u for u in User.objects.all()
    }
    groups = {
        g.name: g for g in Group.objects.all()
    }
    ldap_source = Source.objects.get_or_create(source=SOURCE_LDAP)[0]

    users_by_dn = {}

    for dn, model_attrs in get_ldap_users():
        username = model_attrs['username']
        try:
            user = users.pop(username)
        except KeyError:
            user = User()
        user.is_active = True
        for name, value in model_attrs.items():
            setattr(user, name, value)
        user.save()
        user.sources.add(ldap_source)
        users_by_dn[dn] = user

    for group_attrs in get_ldap_groups():
        name = group_attrs['cn'][0]
        try:
            group = groups.pop(name)
        except KeyError:
            group = Group.objects.create(name=name)
        group.sources.add(ldap_source)
        group.user_set.add(*[
            users_by_dn[dn] for dn in group_attrs.get('member', [])
            if dn in users_by_dn
        ])

    # Remove source ldap users and groups not present in LDAP catalog
    for user in users.values():
        user.sources.remove(ldap_source)
    for group in groups.values():
        group.sources.remove(ldap_source)


def prune_users_and_groups():
    object_names = (
        'inventory_objects',
        'assigned_objects',
        'labs',
        'assigned_labs',
    )

    inactive_users = []
    for user in User.objects.filter(sources=None):
        for obj_name in object_names:
            if hasattr(user, obj_name) and getattr(user, obj_name).all():
                inactive_users.append(user.id)
                log.warn("Inactive LDAP user '%s' has mapped '%s'", user.username, obj_name)
                break
        else:
            log.debug("Delete LDAP user %s from local DB", user.username)
            user.delete()
    User.objects.filter(pk__in=inactive_users).update(is_active=False)

    for group in Group.objects.filter(sources=None):
        for obj_name in object_names:
            if hasattr(group, obj_name) and getattr(group, obj_name).all():
                log.warn("Inactive LDAP group '%s' has mapped '%s'", group.name, obj_name)
                break
        else:
            log.debug("Delete LDAP group %s from local DB", group.name)
            group.delete()
