from django.conf import settings
from django.test import TestCase
from django.contrib.auth.models import User, Group

import dcim.importer
import importers.ldap
import importers.puppet
import importers.switches
import importers.zabbix
import importers.jenkins
from dcim.models import Switch, DCIM_EMPTY, DCIM_SERVER, DCIM_SWITCH, DCIM_OTHER
from inventory.models import (
    FacterLocation,
    InventoryObject,
    InventoryObjectType,
    L2Address,
    Lab,
)
from racks.models import Source, SOURCE_LOCAL, SOURCE_LDAP
from jenkins_ci.models import JenkinsInstance, JenkinsNode, JenkinsLabel

import json
import mockldap
import requests

from datetime import datetime
from mock import patch, call
from pytz import UTC


class ZabbixTestCase(TestCase):
    @patch('pyzabbix.ZabbixAPI')
    def test(self, mock_class):
        instance = mock_class.return_value
        instance.hostgroup.get.return_value = [
            {
                'groupid': 1,
                'name': 'testgroup',
            }
        ]
        instance.host.get.return_value = [
            {
                'host': 'hostname1',
                'hostid': 2,
            }
        ]

        importers.zabbix.import_hosts()

        self.assertItemsEqual(InventoryObject.objects.values('name'), [{'name': 'hostname1'}])


class MockResponse(object):
    pass


class PuppetTestCase(TestCase):
    maxDiff = None

    def setUp(self):
        self.inventory_object = InventoryObject.objects.create(id=1, name='object1')
        self.pk = self.inventory_object.pk
        self.type_unk = InventoryObjectType.objects.create(name='Unknown')
        self.type_vm = InventoryObjectType.objects.create(name='KVM Virtual Machine')
        # This object should be ignored by importer
        InventoryObject.objects.create(id=2, name=None)

    def assert_attributes(self, inventory_object, attributes):
        self.assertItemsEqual(inventory_object.attributes.all().values('name', 'value'), [
            {
                'name': key,
                'value': value,
            }
            for key, value in attributes.items()
        ])

    @patch('requests.get')
    def run_import(self, response, mock_function):
        mock_response = mock_function.return_value = MockResponse()
        mock_response.text = json.dumps(response)

        importers.puppet.import_puppet_attributes()

        self.assertItemsEqual(mock_function.mock_calls, [
            call(u'testpuppet/puppet/node/object1?environment=production',
                 cert=('/etc/ssl/certs/puppet.crt', '/etc/ssl/private/puppet.key'),
                 verify=False,
                 headers={'User-agent': 'Racks Puppet Importer/0.2'})
        ])

    def test_hw_server(self):
        self.inventory_object.update_attributes({
            'to_change': 'original_value',
        })

        self.run_import({
            'data': {
                'parameters': {
                    'virtual': 'physical',
                    'attribute1': 'value1',
                    'mtu_docker0': 'ignored_attribute_value',
                    'to_change': 'changed_value',
                }
            }
        })

        object1 = InventoryObject.objects.get(id=1)

        self.assertEqual(object1.type.name, 'HW Server')
        self.assert_attributes(object1, {
            'attribute1': 'value1',
            'virtual': 'physical',
            'to_change': 'changed_value',
        })

    def test_kvm_server(self):
        self.inventory_object.update_attributes({
            'to_change': 'original_value',
        })

        self.run_import({
            'data': {
                'parameters': {
                    'virtual': 'kvm',
                    'attribute1': 'value1',
                    'mtu_docker0': 'ignored_attribute_value',
                    'to_change': 'changed_value',
                }
            }
        })

        object1 = InventoryObject.objects.get(id=1)

        self.assertEqual(object1.type.name, 'KVM Virtual Machine')
        self.assert_attributes(object1, {
            'attribute1': 'value1',
            'virtual': 'kvm',
            'to_change': 'changed_value',
        })

    def test_keep_object_type_if_set(self):
        self.inventory_object.type = self.type_unk
        self.inventory_object.save()
        self.run_import({
            'data': {
                'parameters': {
                    'virtual': 'kvm',
                }
            }
        })

        # should be set type if type is Unknown
        object1 = InventoryObject.objects.get(id=self.pk)
        self.assertEqual(object1.type, self.type_vm)
        self.run_import({
            'data': {
                'parameters': {
                    'virtual': 'physical',
                }
            }
        })

        # should be kept as VM event if puppet says type HW
        object1 = InventoryObject.objects.get(id=self.pk)
        self.assertEqual(object1.type, self.type_vm)


class DCIMTestCase(TestCase):
    @patch('dcim.importer.Session.request')
    def test_import_servers(self, mock_function):
        mock_function.return_value = [
            {
                'DeviceID': 1,
                'Label': 'label1'
            },
            {
                'DeviceID': 3,
                'Label': 'label3'
            },
        ]

        # This label will be changed)
        InventoryObject.objects.create(id=1, dcim_id=1, dcim_label='oldLabel1')
        # This will be orphaned and pruned, as it's not in DCIM
        InventoryObject.objects.create(id=2, dcim_id=2, dcim_label='oldLabel2')

        # Not orphaned
        # Has name
        InventoryObject.objects.create(id=4, dcim_id=4, dcim_label='label4', name='name4')
        # Has attribute
        obj5 = InventoryObject.objects.create(id=5, dcim_id=5, dcim_label='label5')
        obj5.attributes.create(name='name1', value='value1')
        # Has type
        obj6 = InventoryObject(id=6, dcim_id=6, dcim_label='label6')
        obj6.type = InventoryObjectType.objects.create(name='type1')
        obj6.save()
        # Has location
        obj7 = InventoryObject(id=7, dcim_id=7, dcim_label='label7')
        obj7.facter_location = FacterLocation.objects.create()
        obj7.save()
        # Has label
        obj8 = InventoryObject.objects.create(id=8, dcim_id=8, dcim_label='label8')
        obj8.label.create(name='objectLabel1')
        # Has l2network
        obj9 = InventoryObject.objects.create(id=9, dcim_id=9, dcim_label='label9')
        obj9.l2network.create(name='l2network1', vlanid=1)
        # Has ipv4network
        obj10 = InventoryObject.objects.create(id=10, dcim_id=10, dcim_label='label10')
        obj10.ipv4network.create()
        # Has responsible
        obj11 = InventoryObject.objects.create(id=11, dcim_id=11, dcim_label='label11')
        obj11.responsible.create(username='user1')
        # Has l2address
        obj12 = InventoryObject.objects.create(id=12, dcim_id=12, dcim_label='label12')
        obj12.l2addresses.create(address='112233445566')

        dcim.importer.import_servers()

        self.assertItemsEqual(
            InventoryObject.objects.values_list('id', 'dcim_id', 'dcim_label', 'orphan_dcim_id'),
            [
                (1, 1, 'label1', None),
                (13, 3, 'label3', None),
                # Not pruned
                (4, None, 'label4', 4),
                (5, None, 'label5', 5),
                (6, None, 'label6', 6),
                (7, None, 'label7', 7),
                (8, None, 'label8', 8),
                (9, None, 'label9', 9),
                (10, None, 'label10', 10),
                (11, None, 'label11', 11),
                (12, None, 'label12', 12),
            ])

    @patch('dcim.importer.Session.request')
    def test_import_switches(self, mock_function):
        def side_effect(object_type, object_id=None, filters={}):
            if object_type == 'device':
                return [
                    {
                        'DeviceID': 1,
                        'PrimaryIP': 'ip1',
                        'Label': 'label1',
                    },
                    {
                        'DeviceID': 3,
                        'PrimaryIP': 'ip3',
                        'Label': 'label3',
                    },
                ]

            if object_type == 'deviceport':
                if object_id == 1:
                    return {
                        '1': {
                            'Label': 'portLabel1',
                            'ConnectedDeviceID': None,  # Not connected
                        },
                        '2': {
                            'Label': 'portLabel2',
                            'ConnectedDeviceID': '3',  # Connects to another switch
                        },
                        '3': {
                            'Label': 'portLabel3',
                            'ConnectedDeviceID': '4',  # Connects to server
                        },
                        '4': {
                            'Label': 'portLabel4',
                            'ConnectedDeviceID': '9',  # Connects to unknown device
                        },
                    }

                return {}

        mock_function.side_effect = side_effect

        # Inventory object for port3
        InventoryObject.objects.create(dcim_id=4)

        # This switch will have label changed
        switch1 = Switch.objects.create(dcim_id=1, label='oldLabel1', management_address='oldIp1')

        # These switches will have their connections overwritten with new data
        switch1.ports.create(number=2, label='portOldLabel2')
        switch1.ports.create(
            number=3, label='portLabel3', dcim_server_id=7, dcim_device_type=DCIM_SWITCH)
        switch1.ports.create(
            number=4, label='portLabel4', dcim_server_id=8, dcim_device_type=DCIM_SERVER)

        # This port will get deleted
        switch1.ports.create(number=5, label='portLabel5')

        # This switch will get deleted
        Switch.objects.create(dcim_id=2, label='label2'),

        dcim.importer.import_switches()

        self.assertItemsEqual(
            Switch.objects.values('label', 'management_address'),
            [
                {'label': 'label1', 'management_address': 'ip1'},
                {'label': 'label3', 'management_address': 'ip3'},
            ]
        )

        resulting_ports = Switch.objects.get(label='label1').ports.values(
            'number', 'label', 'dcim_server_id', 'dcim_device_type'
        )
        self.assertItemsEqual(
            resulting_ports,
            [
                # New port, no DCIM connection
                {'number': '1', 'label': 'portLabel1',
                 'dcim_server_id': None, 'dcim_device_type': DCIM_EMPTY},
                # Label changed, DCIM connection to Switch #3
                {'number': '2', 'label': 'portLabel2',
                 'dcim_server_id': 3, 'dcim_device_type': DCIM_SWITCH},
                # DCIM connection changed to Server #4
                {'number': '3', 'label': 'portLabel3',
                 'dcim_server_id': 4, 'dcim_device_type': DCIM_SERVER},
                # DCIM connection changed to Other #9
                {'number': '4', 'label': 'portLabel4',
                 'dcim_server_id': 9, 'dcim_device_type': DCIM_OTHER},
            ]
        )


class SwitchesTestCase(TestCase):
    @patch('importers.switches.load_credentials')
    @patch('importers.switches.mac_to_port')
    def test_import_address_tables(self, mock_mac_to_port, mock_load_credentials):
        mock_mac_to_port.return_value = [
            # invalid address, won't get added
            ('xxxxx', 'Gi0/1'),
            ('aaaaaa000001', 'Gi0/2'),
        ]

        Switch.objects.create(dcim_id=1, label='ip1', management_address='ip1')

        importers.switches.import_address_tables()

        self.assertItemsEqual(
            L2Address.objects.values('address'),
            [
                {'address': 'aaaaaa000001'},
            ]
        )

        addr1 = L2Address.objects.get(address='aaaaaa000001')

        self.assertItemsEqual(addr1.switch_ports.values_list('switch_port_name'), [('Gi0/2',)])

    def test_bind_dcim_to_switch_ports(self):
        switch1 = Switch.objects.create(dcim_id=1, label='ip1', management_address='ip1')

        # This should stay intact
        switch1.ports.create(id=1, label='GigabitEthernet0/1', switch_port_name='Gi0/1')
        port2 = switch1.ports.create(id=2, label='GigabitEthernet0/2')
        port2.l2addresses.create(address='aaaaaa000001')
        # This should get merged to port2
        port3 = switch1.ports.create(id=3, switch_port_name='Gi0/2')
        port3.l2addresses.create(address='aaaaaa000002')

        importers.switches.bind_dcim_to_switch_ports()

        self.assertItemsEqual(switch1.ports.values_list('id', 'label', 'switch_port_name'), [
            (1, 'GigabitEthernet0/1', 'Gi0/1'),
            (2, 'GigabitEthernet0/2', 'Gi0/2'),
        ])

        self.assertItemsEqual(port2.l2addresses.values_list('address'), [
            ('aaaaaa000001',),
            ('aaaaaa000002',),
        ])

    def test_bind_dcim_servers_to_inventory(self):
        obj1 = InventoryObject.objects.create(name='obj1')
        # This will get merged into obj1 and pruned
        obj2 = InventoryObject.objects.create(dcim_id=2)

        l2address1 = obj1.l2addresses.create(address='aaaaaa000001')

        switch3 = Switch.objects.create(dcim_id=3)
        port1 = switch3.ports.create(dcim_server_id=obj2.dcim_id, dcim_device_type=DCIM_SERVER)

        port1.l2addresses.add(l2address1)

        importers.switches.bind_dcim_servers_to_inventory()

        self.assertItemsEqual(
            InventoryObject.objects.values_list('name', 'dcim_id'),
            [
                # obj2 bound to obj1
                ('obj1', 2),
            ]
        )


class LDAPImportTestCase(TestCase):
    DIRECTORY = {
        settings.LDAP_USER_SEARCH.base_dn: {
            'ou': ['people']
        },
        settings.AUTH_LDAP_USER_DN_TEMPLATE % {'user': 'user1'}: {
            'uid': ['user1'],
            'objectClass': ['person'],
            'givenName': ['Given1'],
            'mail': ['user1@mail'],
            'sn': ['sn1'],
        },
        settings.AUTH_LDAP_GROUP_SEARCH.base_dn: {
            'cn': ['group1'],
            'objectClass': ['groupOfNames'],
            'member': [settings.AUTH_LDAP_USER_DN_TEMPLATE % {'user': 'user1'}],
        }
    }

    @classmethod
    def setUpClass(cls):
        cls.mockldap = mockldap.MockLdap(cls.DIRECTORY)

    @classmethod
    def tearDownClass(cls):
        del cls.mockldap

    def setUp(self):
        self.mockldap.start()
        self.ldapobj = self.mockldap[settings.AUTH_LDAP_SERVER_URI]

    def tearDown(self):
        # Stop patching ldap.initialize and reset state.
        self.mockldap.stop()
        del self.ldapobj

    def test_import_users(self):
        local_source = Source.objects.get_or_create(source=SOURCE_LOCAL)[0]
        ldap_source = Source.objects.get_or_create(source=SOURCE_LDAP)[0]

        user_counter = len(User.objects.all())

        # These user and group should be deleted by the end: they only have LDAP source, are not
        # present in mocked LDAP response and don't have any assigned/responsible objects.
        User.objects.create(username='user2').sources.add(ldap_source)
        Group.objects.create(name='group2').sources.add(ldap_source)

        # This users should be marked as deleted
        responsible_object_user = User.objects.create(username='responsible_object_user')
        inventory_object = InventoryObject.objects.create()
        inventory_object.responsible.add(responsible_object_user)
        inventory_object.assigned_user = User.objects.create(username='assigned_object_user')

        responsible_lab_user = User.objects.create(username='responsible_lab_user')
        lab = Lab.objects.create()
        lab.responsible.add(responsible_lab_user)
        lab.assigned_user = User.objects.create(username='assigned_lab_user')
        lab.assigned_user.sources.add(ldap_source)

        # This groups should be marked deleted
        inventory_object.assigned_group = Group.objects.create(name='assigned_object_group')
        lab.assigned_group = Group.objects.create(name='assigned_lab_group')

        inventory_object.save()
        lab.save()

        User.objects.create(username='unrelated_user').sources.add(ldap_source, local_source)

        local_source.users.add(*User.objects.filter(sources=None))
        local_source.groups.add(*Group.objects.filter(sources=None))

        importers.ldap.import_users()

        self.assertEqual(User.objects.count(), user_counter + 7)
        importers.ldap.prune_users_and_groups()
        self.assertEqual(User.objects.count(), user_counter + 6)

        user1 = User.objects.get(username='user1')

        self.assertEqual(user1.first_name, 'Given1')
        self.assertEqual(user1.last_name, 'sn1')
        self.assertEqual(user1.email, 'user1@mail')
        self.assertEqual([g.name for g in user1.groups.all()], ['group1'])

        unrelated = User.objects.get(username='unrelated_user')

        self.assertIn(local_source, unrelated.sources.all())
        self.assertNotIn(ldap_source, unrelated.sources.all())

        deactivated_user = User.objects.get(username='assigned_lab_user')

        # This user only had SOURCE_LDAP, was not present in mocked LDAP response and has assigned
        # objects/labs. It must have no sources remaining and should be marked as inactive.
        self.assertFalse(deactivated_user.sources.all())
        self.assertFalse(deactivated_user.is_active)


BASE_TIMESTAMP = 1400000000


class MockAPIError(Exception):
    pass


def jenkins_api_mock(url, *args, **kwargs):

    class MockAPIResponse(object):
        status_code = requests.codes.ok
        text = None
        parsed_json = None

        def __init__(self, status_code, text):
            self.status_code = status_code
            if isinstance(text, (str, unicode)):
                self.text = text
            else:
                self.text = json.dumps(text)
                self.parsed_json = text

        def json(self):
            return self.parsed_json

    VALID_REQUESTS = {

        # Main test case (correct import)

        'http://jenkins/computer/api/json': {
            'computer': [
                {'displayName': 'master'},
                {'displayName': 'slave_a', 'offline': False},
                {'displayName': 'slave_b', 'offline': False},
                {'displayName': 'slave_e', 'offline': False},
            ]
        },

        'http://jenkins/api/json?tree=jobs[lastBuild[timestamp,builtOn]]': {
            'jobs': [
                {'lastBuild': {'timestamp': BASE_TIMESTAMP * 1000, 'builtOn': 'slave_a'}},
                {'lastBuild': {'timestamp': BASE_TIMESTAMP * 1000, 'builtOn': 'slave_b'}},
            ]
        },

        'http://jenkins/computer/(master)/config.xml': '''<?xml version="1.0" encoding="UTF-8"?>
            <hudson><label></label></hudson>
        ''',

        'http://jenkins/computer/slave_a/config.xml': '''<?xml version="1.0" encoding="UTF-8"?>
            <slave>
                <name>slave_a</name>
                <launcher><host>www.google.com</host></launcher>
                <label>label1 label2</label>
            </slave>
        ''',

        'http://jenkins/computer/slave_b/config.xml': '''<?xml version="1.0" encoding="UTF-8"?>
            <slave>
                <name>slave_b</name>
                <launcher><host>www.amazon.com</host></launcher>
                <label>label2 label3</label>
            </slave>
        ''',

        'http://jenkins/computer/slave_e/config.xml': '''<?xml version="1.0" encoding="UTF-8"?>
            <slave><name>slave_e</name></slave>
        ''',

        # Secondary test case (import with errors)

        'http://jenkins2/computer/api/json': {
            'computer': [
                {'displayName': 'master'},
                {'displayName': 'slave_x', 'offline': False},
            ]
        },

        'http://jenkins2/api/json?tree=jobs[lastBuild[timestamp,builtOn]]': {'jobs': []},

        'http://jenkins2/computer/(master)/config.xml': requests.codes.forbidden,

        'http://jenkins2/computer/slave_x/config.xml': MockAPIError,

        # Tertiary test case (failed import from YAML)

        'http://jenkins3/computer/api/json': requests.codes.not_found,

    }

    try:
        response = VALID_REQUESTS[url]
    except KeyError:
        return MockAPIResponse(requests.codes.not_found, '')
    if isinstance(response, int):
        return MockAPIResponse(response, '')
    elif isinstance(response, type) and issubclass(response, Exception):
        raise response()
    else:
        return MockAPIResponse(requests.codes.ok, response)


class JenkinsImportTestCase(TestCase):

    def test_jenkins_binding_to_inventory_objects(self):

        invobj_a = InventoryObject.objects.create(name='www.google.com')
        invobj_b = InventoryObject.objects.create(name='www.amazon.com')

        JenkinsNode.objects.bulk_create([
            JenkinsNode(name='slave_a', hostname='www.google.com'),
            JenkinsNode(name='slave_b', hostname='www.amazon.com'),
            JenkinsNode(name='slave_c', hostname='www.amazon.com', inventory_object=invobj_b),
            JenkinsNode(name='slave_d', hostname='www.orphan.com'),
        ])

        importers.jenkins.bind_jenkins_nodes_to_inventory()

        # Refresh information from DB
        slave_a = JenkinsNode.objects.get(name='slave_a')
        slave_b = JenkinsNode.objects.get(name='slave_b')
        slave_c = JenkinsNode.objects.get(name='slave_c')
        slave_d = JenkinsNode.objects.get(name='slave_d')

        # Test Case 1. Unbound Node correctly binds to InvObj via hostname==name match.
        self.assertEqual(slave_a.inventory_object, invobj_a)
        self.assertEqual(slave_b.inventory_object, invobj_b)

        # Test Case 2. Bound Node is not affected even when there's an obvious match.
        self.assertEqual(slave_c.inventory_object, invobj_b)

        # Test Case 3. Several Nodes can be successfully bound to a single InvObj.
        self.assertEqual(slave_b.inventory_object, slave_c.inventory_object)

        # Test Case 4. Node with no matches is not affected.
        self.assertIsNone(slave_d.inventory_object)

    @patch('importers.jenkins.load_credentials')
    @patch('requests.get')
    def test_jenkins_import(self, mock_requests_get, mock_load_credentials):

        mock_requests_get.side_effect = jenkins_api_mock
        mock_load_credentials.return_value = [
            {'name': 'Updated', 'url': 'http://jenkins/', 'username': 'test', 'api_key': 'test'},
            {'name': 'CI404', 'url': 'http://jenkins3/', 'username': 'test', 'api_key': 'test'},
        ]

        # Initialize test inventory object and Jenkins instance
        invobj_a = InventoryObject.objects.create(name='www.google.com')

        test_instance = JenkinsInstance.objects.create(
            name='Test', url='http://jenkins/', username='fake', api_key='fake'
        )
        fail_instance = JenkinsInstance.objects.create(
            name='Fail', url='http://jenkins2/', username='test', api_key='test'
        )
        ignore_instance = JenkinsInstance.objects.create(
            name='Ignore', url='http://jenkins4/', username=None, api_key=None
        )

        # Preinitialize Jenkins nodes in DB
        slave_a = JenkinsNode.objects.create(
            instance=test_instance,
            name='slave_a',
            last_job=datetime.utcfromtimestamp(BASE_TIMESTAMP + 1000).replace(tzinfo=UTC)
        )
        slave_b = JenkinsNode.objects.create(
            name='slave_b',
            instance=test_instance,
            hostname='www.amazon.com',
            last_job=datetime.utcfromtimestamp(BASE_TIMESTAMP - 1000).replace(tzinfo=UTC),
            disabled=True,
        )
        slave_c = JenkinsNode.objects.create(
            name='slave_c',
            instance=test_instance,
            inventory_object=invobj_a,
        )
        slave_d = JenkinsNode.objects.create(
            name='slave_d',
            instance=test_instance,
        )

        # Run importer with mocked `requests.get`
        importers.jenkins.import_all_jenkins()

        # Refresh information from DB
        test_instance = JenkinsInstance.objects.get(pk=test_instance.pk)
        ignore_instance = JenkinsInstance.objects.get(pk=ignore_instance.pk)
        ci404_instance = JenkinsInstance.objects.get(name='CI404')

        slave_a = JenkinsNode.objects.get(pk=slave_a.pk)
        slave_b = JenkinsNode.objects.get(pk=slave_b.pk)
        slave_e = JenkinsNode.objects.get(name='slave_e')

        label1 = JenkinsLabel.objects.get(name='label1')
        label2 = JenkinsLabel.objects.get(name='label2')
        label3 = JenkinsLabel.objects.get(name='label3')

        # Validate test cases

        # Check #0. Check how instances interacted with YAML settings.
        self.assertEqual(test_instance.name, 'Updated')  # name was updated from YAML
        self.assertEqual(test_instance.username, 'fake')  # username was not updated
        self.assertIsNone(ignore_instance.last_failed_sync)  # was not processed at all
        self.assertIsNone(ignore_instance.last_successful_sync)
        self.assertIsNotNone(ci404_instance.last_failed_sync)  # YAML CI exists and was processed

        mock_requests_get.assert_any_call('http://jenkins/computer/api/json', auth=('test', 'test'))

        # Check #1. Ensure slaves have been assigned correctly.
        self.assertEqual(test_instance, slave_a.instance)  # Found by instance+name and kept
        self.assertEqual(test_instance, slave_b.instance)  # Orphan found by hostname+IP
        self.assertEqual(test_instance, slave_e.instance)  # New instance created from API response

        # Check #2. Ensure orphaned associated slave is deleted from database.
        self.assertRaises(JenkinsNode.DoesNotExist, JenkinsNode.objects.get, pk=slave_c.pk)

        # Check #3. Ensure orphaned unassociated slave is deleted from database.
        self.assertRaises(JenkinsNode.DoesNotExist, JenkinsNode.objects.get, pk=slave_d.pk)

        # Check #4. Ensure labels were imported and assigned correctly.
        self.assertIn(label1, slave_a.labels.all())
        self.assertIn(label2, slave_a.labels.all())
        self.assertIn(label2, slave_b.labels.all())
        self.assertIn(label3, slave_b.labels.all())
        self.assertNotIn(label3, slave_a.labels.all())
        self.assertNotIn(label1, slave_b.labels.all())

        # Check #5. Ensure hostnames were imported and assigned correctly
        self.assertEqual(slave_a.hostname, 'www.google.com')
        self.assertEqual(slave_b.hostname, 'www.amazon.com')

        # Check #6. Ensure last_job has been updated correctly from mocked response
        higher_datetime = datetime.utcfromtimestamp(BASE_TIMESTAMP + 1000).replace(tzinfo=UTC)
        lower_datetime = datetime.utcfromtimestamp(BASE_TIMESTAMP).replace(tzinfo=UTC)
        self.assertEqual(slave_a.last_job, higher_datetime)  # was BASE+1000, got BASE from xml
        self.assertEqual(slave_b.last_job, lower_datetime)  # was BASE-1000, got BASE from xml

        # Check #7. Ensure secondary Jenkins instance is marked as failed.
        fail_instance = JenkinsInstance.objects.get(pk=fail_instance.pk)
        self.assertIsNone(fail_instance.last_successful_sync)
        self.assertIsNotNone(fail_instance.last_failed_sync)

        # Check #8. Ensure that slaves have been loaded despite failed sync.
        slave_x = JenkinsNode.objects.get(name='slave_x')
        self.assertEqual(slave_x.instance, fail_instance)
