from django import forms
from django.db import models
from django.utils.six import with_metaclass

from racks.encryption import enigma, EncryptionError


class NullableCharFieldFormField(forms.CharField):

    def to_python(self, *args, **kwargs):
        value = super(NullableCharFieldFormField, self).to_python(*args, **kwargs)
        if value == '':
            return None
        return value


class NullableCharField(models.CharField):
    def formfield(self, **kwargs):
        if 'form_class' not in kwargs:
            kwargs['form_class'] = NullableCharFieldFormField
        return super(NullableCharField, self).formfield(**kwargs)


class EncryptedCharField(with_metaclass(models.SubfieldBase, models.CharField)):

    def __init__(self, *args, **kwargs):
        kwargs['max_length'] = 255
        super(EncryptedCharField, self).__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super(EncryptedCharField, self).deconstruct()
        del kwargs['max_length']
        return name, path, args, kwargs

    def get_internal_type(self):
        return 'CharField'

    def to_python(self, value):
        if value is None or len(value) <= 64:
            return value
        try:
            return enigma.decrypt(value)
        except EncryptionError:
            return None

    def get_prep_value(self, value):
        return enigma.encrypt(value)

    def get_prep_lookup(self, lookup_type, value):
        if lookup_type in ('exact', 'isnull'):
            return self.get_prep_value(value)
        else:
            raise TypeError(
                'lookup type %r not supported by %s' % (lookup_type, type(self).__name__)
            )

    def formfield(self, **kwargs):
        if 'form_class' not in kwargs:
            kwargs['form_class'] = forms.CharField
        kwargs['max_length'] = 64
        return super(EncryptedCharField, self).formfield(**kwargs)


# `south` is not installed for 1.8, making this code optional
try:
    from south.modelsinspector import add_introspection_rules
    add_introspection_rules([], [
        r"^inventory\.fields\.NullableCharField",
        r"^inventory\.fields\.EncryptedCharField",
    ])
except ImportError:
    pass
