import netaddr
import re
import urlparse
import logging

from collections import OrderedDict

from django.conf import settings
from django.contrib.auth.models import User, Group
from django.core.urlresolvers import reverse
from django.core.exceptions import ValidationError
from django.core.validators import MaxValueValidator, MinValueValidator, URLValidator
from django.db import models
from django.utils.http import urlencode
from django.utils.functional import cached_property
from django.utils.translation import ugettext_lazy as _

from inventory.fields import NullableCharField, EncryptedCharField
from security import permissions
from racks.models import RacksModel, RacksTimestampedModel

logger = logging.getLogger(__name__)


def validate_hostname(value):
    if not value:
        return value
    if ('/' in value) or (':' in value):
        raise ValidationError(_('Please only enter hostname: "www.google.com" is ok, but '
                                '"http://www.google.com" or "www.google.com/search" are not'))
    validator = URLValidator()
    try:
        validator('http://{}'.format(value))
    except ValidationError:
        if '.' in value:
            raise ValidationError(_('Please enter a valid hostname'))
        try:
            validator('http://{}.com'.format(value))
        except ValidationError:
            raise ValidationError(_('Please enter a valid hostname'))


class Location(RacksTimestampedModel):
    name = models.CharField(max_length=254, unique=True)

    def __unicode__(self):
        return self.name

    class Meta:
        ordering = ('name',)


class FacterLocation(RacksTimestampedModel):
    fact = models.CharField(max_length=32, unique=True)
    location = models.ForeignKey('Location', on_delete=models.SET_NULL, related_name='facts',
                                 null=True, blank=True)

    def __unicode__(self):
        return "{} ({})".format(self.fact, self.location)


class L2Network(RacksTimestampedModel):
    name = models.CharField(max_length=254)
    vlanid = models.IntegerField(validators=[
        MinValueValidator(1),
        MaxValueValidator(4094),
    ], verbose_name='VLAN ID')

    class Meta:
        verbose_name = 'VLAN'

    def __unicode__(self):
        if self.name:
            return u'{} (VLAN ID: {})'.format(self.name, self.vlanid)

        return u"VLAN ID: {}".format(self.vlanid)


class L2Address(RacksModel):
    MAC_ADDRESS_RE = re.compile(r'^[0-9a-f]{12}$')

    address = models.CharField(max_length=12, unique=True)
    inventory_object = models.ForeignKey('inventory.InventoryObject', null=True,
                                         related_name='l2addresses')

    @classmethod
    def is_valid_mac_address(cls, mac_address):
        return cls.MAC_ADDRESS_RE.match(mac_address)

    def __unicode__(self):
        return u":".join(self.address[i:i + 2] for i in range(0, len(self.address), 2)).upper()


class IPv4Network(RacksTimestampedModel):
    address = models.CharField(max_length=15)
    netmask = models.CharField(max_length=15)
    gateway = models.CharField(max_length=15)
    l2network = models.ForeignKey(
        L2Network, on_delete=models.SET_NULL, related_name='ipv4_networks', null=True, blank=True)

    class Meta:
        unique_together = ('address', 'netmask')
        ordering = ('address',)
        verbose_name = 'IPv4 Network'

    def __str__(self):
        return str(netaddr.IPNetwork('%s/%s' % (self.address, self.netmask)))

    def __unicode__(self):
        return unicode(netaddr.IPNetwork('%s/%s' % (self.address, self.netmask)))


class InventoryObjectType(RacksModel):
    name = models.CharField(max_length=254, unique=True)

    def __str__(self):
        return self.name

    def __unicode__(self):
        return self.name


class InventoryObjectLabel(RacksTimestampedModel):
    TYPE_CHOICES = (
        ('GY', 'Grey'),
        ('BL', 'Blue'),
        ('GN', 'Green'),
        ('LB', 'Lightblue'),
        ('OR', 'Orange'),
        ('RD', 'Red'),
    )

    name = models.CharField(max_length=64, unique=True)
    type = models.CharField(
        max_length=2,
        choices=TYPE_CHOICES,
        default='GY'
    )

    def __unicode__(self):
        return self.name

    def __str__(self):
        return self.name

    def get_bootstrap_color(self):
        colors = {
            'GY': 'default',
            'BL': 'primary',
            'GN': 'success',
            'LB': 'info',
            'OR': 'warning',
            'RD': 'danger',
        }

        return colors.get(self.type, 'default')

    def get_absolute_url(self):
        return reverse('inventory.objects.list') + '?' + urlencode({
            'label': self.name
        })


class InventoryObject(RacksTimestampedModel):
    name = NullableCharField(max_length=254, unique=True, db_index=True, null=True, blank=True,
                             verbose_name='hostname', validators=[validate_hostname])
    type = models.ForeignKey(InventoryObjectType, related_name='inventory_objects', null=True,
                             blank=True, verbose_name='inventory object type')
    facter_location = models.ForeignKey('FacterLocation', related_name='inventory_objects',
                                        null=True, blank=True)
    label = models.ManyToManyField(InventoryObjectLabel, related_name='inventory_objects',
                                   blank=True)
    l2network = models.ManyToManyField(L2Network, related_name='inventory_objects', blank=True,
                                       verbose_name='VLAN')
    ipv4network = models.ManyToManyField(IPv4Network, related_name='inventory_objects', blank=True,
                                         verbose_name='IPv4 network')
    responsible = models.ManyToManyField(User, related_name='inventory_objects',
                                         db_index=True, blank=True)

    ipmi_host = NullableCharField(max_length=255, unique=True, null=True, blank=True,
                                  verbose_name='IPMI hostname')
    ipmi_user = NullableCharField(max_length=255, null=True, blank=True,
                                  verbose_name='IPMI username')
    ipmi_password = EncryptedCharField(max_length=255, null=True, blank=True,
                                       verbose_name='IPMI password')

    dcim_id = models.IntegerField(null=True, db_index=True, unique=True, blank=True,
                                  verbose_name='DCIM ID')
    orphan_dcim_id = models.IntegerField(null=True, db_index=True, blank=True,
                                         verbose_name='orphan DCIM ID')
    dcim_label = NullableCharField(max_length=255, null=True, blank=True,
                                   verbose_name='DCIM label')
    dcim_jira_issue = models.CharField(max_length=255, null=True, blank=True,
                                       verbose_name='DCIM Jira issue')

    bugs = models.CommaSeparatedIntegerField(blank=True, null=True, max_length=254)

    assigned_user = models.ForeignKey(User, blank=True, null=True,
                                      related_name='assigned_objects')
    assigned_group = models.ForeignKey(Group, blank=True, null=True,
                                       related_name='assigned_objects')

    class Meta:
        ordering = ('-id', )

    def __unicode__(self):
        return "{} ({})".format(self.name, self.dcim_label)

    def update_attributes(self, attributes):
        for name, value in attributes.items():
            attribute_object, created = self.attributes.get_or_create(name=name,
                                                                      defaults={'value': value})

            if not created and attribute_object.value != value:
                attribute_object.value = value
                attribute_object.save()

    @cached_property
    def all_attributes(self):
        return dict((attr.name, attr.value) for attr in self.attributes.all())

    @property
    def cpu(self):
        try:
            return '{processorcount}x{processor0}'.format(**self.all_attributes)
        except KeyError:
            return None

    @property
    def display_name(self):
        if self.name:
            return self.name
        elif self.dcim_label:
            return 'DCIM Entry: %s' % (self.dcim_label, )
        elif self.dcim_id:
            return 'DCIM Entry ID: %s' % (self.dcim_id, )
        elif self.orphan_dcim_id:
            return 'Orphaned DCIM Entry ID: %s' % (self.orphan_dcim_id, )
        else:
            return 'Empty object ID: %s' % (self.id, )

    def dump(self):
        return {
            'name': self.name,
            'hardware': self.hardware,
        }

    @property
    def hardware(self):
        return self.all_attributes.get('hardware', None)

    @property
    def ipmi_link(self):
        try:
            ipmi_host = self.all_attributes['ipmi_host']
        except KeyError:
            return None
        if urlparse.urlparse(ipmi_host).scheme != '':
            return ipmi_host
        return 'http://%s' % self.ipmi_host

    @property
    def ipv4address(self):
        return self.all_attributes.get('ipaddress', None)

    @property
    def ipv4networks(self):
        return [
            '%s/%s' % (x.address, x.netmask) for x in self.ipv4network.all()]

    @property
    def jira_issue(self):
        try:
            return '{0}{jira_issue}'.format(settings.JIRA_BROWSE_URL, **self.all_attributes)
        except KeyError:
            return None

    @property
    def dcim_jira_issue_link(self):
        if self.dcim_jira_issue:
            return settings.JIRA_BROWSE_URL + self.dcim_jira_issue

    @property
    def macaddress(self):
        result = []
        for key, value in self.all_attributes.iteritems():
            if 'macaddress' in key:
                result.append(value)
        return result

    @property
    def memory(self):
        return self.all_attributes.get('memorysize', None)

    @property
    def ssh_link(self):
        try:
            return 'ssh://{fqdn}'.format(**self.all_attributes)
        except KeyError:
            return None

    def get_absolute_url(self):
        return reverse('inventory.objects.detail', args=(self.pk,))

    @property
    def label_string(self):
        return ','.join([x.name for x in self.label.all()])

    @property
    def macaddress_string(self):
        return ','.join(self.macaddress)

    @property
    def responsible_string(self):
        return ','.join([x.username for x in self.responsible.all()])

    @property
    def assigned_user_name(self):
        if self.assigned_user is None:
            return u''
        return self.assigned_user.username

    @property
    def assigned_group_name(self):
        if self.assigned_group is None:
            return u''
        return self.assigned_group.name

    @property
    def zabbix_link(self):
        url = 'https://monitoring.infra.mirantis.net/hostinventories.php?hostid={zabbix_host_id}'
        try:
            return url.format(**self.all_attributes)
        except KeyError:
            return None

    @property
    def dcim_link(self):
        if self.dcim_id:
            return settings.DCIM_URL + 'devices.php?' + urlencode({'DeviceID': self.dcim_id})

    @property
    def location_name(self):
        if self.facter_location and self.facter_location.location:
            return self.facter_location.location.name

        return None

    @property
    def location_fact(self):
        if self.facter_location:
            return self.facter_location.fact

        return None

    def dcim_orphan(self):
        self.orphan_dcim_id = self.dcim_id
        self.dcim_id = None
        self.save()

    @classmethod
    def prune_orphaned_objects(cls):
        """Remove object which were merged into other and have no related entries"""

        orphaned_objects = cls.objects.filter(name=None,
                                              type=None,
                                              facter_location=None,
                                              label=None,
                                              l2network=None,
                                              ipv4network=None,
                                              responsible=None,
                                              attributes=None,
                                              l2addresses=None,
                                              dcim_id=None)

        orphaned_objects.delete()

    @property
    def bugs_list(self):
        return self.bugs.split(',') if self.bugs else []

    @property
    def used_in_string(self):
        distinct_cis = {node.instance for node in self.jenkins_nodes.all() if node.instance}
        allocs = [instance.name for instance in distinct_cis]
        allocs.extend([lab.name for lab in self.labs.all()])
        return ','.join(sorted(allocs))

    SEARCHABLE_FIELDS_MAPPING = OrderedDict({
        'Name': 'name',
        'Location': 'facter_location__location__name',
        'Facter Location': 'facter_location__fact',
        'DCIM Label': 'dcim_label',
        'Type': 'type__name',
        'IPMI Host': 'ipmi_host',
        'IPMI User': 'ipmi_user',
        'DCIM Jira Issue': 'dcim_jira_issue',
        'Assigned User Username': 'assigned_user__username',
        'Assigned User First Name': 'assigned_user__first_name',
        'Assigned User Last Name': 'assigned_user__last_name',
        'Assigned User Email': 'assigned_user__email',
        'Assigned Group': 'assigned_group__name',
    })

    PUBLIC_FIELDS = (
        'id',
        'created',
        'modified',
        'name',
        'type',
        'facter_location',
        'label',
        'l2network',
        'ipv4network',
        'responsible',
        # 'ipmi_host',
        # 'ipmi_user',
        # 'ipmi_password',
        'dcim_id',
        'orphan_dcim_id',
        'dcim_label',
        'bugs',
        'assigned_user',
        'assigned_group',
        'cpu',
        'display_name',
        'hardware',
        # 'ipmi_link',
        'ipv4address',
        'ipv4networks',
        'jira_issue',
        'dcim_jira_issue',
        'macaddress',
        'memory',
        'ssh_link',
        'zabbix_link',
        'dcim_link',
        'location_name',
        'location_fact',
        'bugs_list',
        'attributes',
        'l2addresses',
        'jenkins_nodes',
    )

    @classmethod
    def get_permissions(cls, user, obj):
        if obj is not None and user is not None and obj.can_view_private(user):
            return permissions.VIEW_ALL

        return permissions.Permissions(view=cls.PUBLIC_FIELDS)

    def can_view_private(self, user):
        if user:
            if self.assigned_user == user:
                return True

            if self.assigned_group in user.groups.all():
                return True

            for lab in self.labs.all():
                if lab.assigned_user == user:
                    return True

                if lab.assigned_group in user.groups.all():
                    return True

        return False

    @classmethod
    def get_queryset_for_permissions(cls, qs, user):
        permissions_by_object = {}

        for obj in qs.prefetch_related('assigned_user', 'assigned_group', 'labs',
                                       'labs__assigned_user', 'labs__assigned_group'):
            permissions_by_object[obj] = obj.get_permissions(user)

        return permissions_by_object


class InventoryObjectCredential(RacksTimestampedModel):
    name = models.CharField(max_length=255)
    context = models.CharField(max_length=255)
    value = models.CharField(max_length=255)
    inventory_object = models.ForeignKey(
        InventoryObject, related_name='inventory_objects')

    class Meta:
        unique_together = ('name', 'context', 'inventory_object')

    def __unicode__(self):
        return '%s for %s@%s' % (
            self.name, self.context, self.inventory_object.name)


class InventoryObjectAttribute(RacksTimestampedModel):
    name = models.CharField(max_length=254)
    value = models.TextField()
    object = models.ForeignKey(InventoryObject, related_name='attributes')

    class Meta:
        unique_together = ('name', 'object')
        ordering = ('name',)

    def __unicode__(self):
        return '%s@%s' % (self.name, self.object.name)

    def __str__(self):
        return '%s@%s' % (self.name, self.object.name)

    def awesome_value(self):
        handlers = [
            'fqdn',
            'version',
            'zabbix_host_id',
        ]
        if self.name in handlers:
            if self.name == 'zabbix_host_id':
                return '<a href="%s" target="_blank">%s</a>' % (
                    self.object.zabbix_link,
                    self.value
                )
            if self.name == 'version':
                sha1commit = re.search(r'([a-f0-9]{40})', self.value)
                if sha1commit and len(sha1commit.groups()) > 0:
                    try:
                        return '<a href="%s" target="_blank">%s</a>' % (
                            settings.COMMIT_HASH_URL_PATTERN % (sha1commit.groups()[0]),
                            self.value,
                        )
                    except AttributeError:
                        logger.warn("Missing COMMIT_HASH_URL_PATTERN in settings")
                return self.value
            if self.name == 'fqdn':
                return '<a href="ssh://%s">%s</a>' % (
                    self.value,
                    self.value,
                )

        return self.value


class Lab(RacksTimestampedModel):
    name = models.CharField(max_length=255, blank=False)
    vlans = models.ManyToManyField('L2Network', blank=True, verbose_name='VLAN')
    responsible = models.ManyToManyField(User, blank=True, related_name='labs')
    inventory_objects = models.ManyToManyField('InventoryObject', related_name='labs', blank=True)
    bugs = models.CommaSeparatedIntegerField(blank=True, null=True, max_length=254)
    assigned_user = models.ForeignKey(User, blank=True, null=True,
                                      related_name='assigned_labs')
    assigned_group = models.ForeignKey(Group, blank=True, null=True,
                                       related_name='assigned_labs')

    @property
    def bugs_list(self):
        return self.bugs.split(',') if self.bugs else []
