from __future__ import unicode_literals

import unittest
import json

from django.contrib.auth.models import User
from mock import patch
from rest_framework.test import APIClient, APITestCase

from inventory.models import (
    InventoryObject,
    InventoryObjectLabel,
    InventoryObjectType,
    IPv4Network,
    Location,
)
from security.models import Application
from security.permissions import Permissions


@unittest.skip("All API tests temporarily disabled")
class APIv2TestCase(APITestCase):
    maxDiff = None

    # Primary keys for test fixtures (used in URLs and some asserts)
    user_pk = None
    loc_pk = None
    fact1_pk = None
    fact2_pk = None
    vlan_pk = None
    type_pk = None
    label_pk = None
    obj1_pk = None
    obj2_pk = None
    attr1_pk = None
    attr2_pk = None
    attr3_pk = None

    def setUp(self):
        user1 = User.objects.create_superuser(
            username='user1', password='password1', email='user1@mail'
        )
        self.user_pk = user1.pk

        self.client.force_authenticate(user=user1)

        # TODO: group serialization does not work yet
        # group1 = Group.objects.create()
        # user1.groups.add(group1)

        Application.objects.create(user=user1, name='authapp1',
                                   auth_token='authtoken1', active=True)

        Application.objects.create(user=user1, name='authapp2',
                                   auth_token='authtoken2', active=True)

        location1 = Location.objects.create(name='location1_name')
        self.loc_pk = location1.pk

        facter_location1 = location1.facts.create(
            fact='facter_location1_name',
        )
        facter_location2 = location1.facts.create(
            fact='facter_location2_name',
        )
        self.fact1_pk = facter_location1.pk
        self.fact2_pk = facter_location2.pk

        self.vlan_pk = IPv4Network.objects.create(
            address='10.0.0.1', netmask='255.0.0.0', gateway='10.0.0.254'
        ).pk

        type1 = InventoryObjectType.objects.create(name='type1')
        self.type_pk = type1.pk

        label1 = InventoryObjectLabel.objects.create(name='label1')
        self.label_pk = label1.pk

        obj1 = InventoryObject.objects.create(
            name='object1.com', facter_location=facter_location1, type=type1,
        )
        self.obj1_pk = obj1.pk

        self.attr1_pk = obj1.attributes.create(name='attribute1', value='obj1 attribute1 value').pk
        self.attr2_pk = obj1.attributes.create(name='attribute2', value='obj1 attribute2 value').pk

        obj2 = InventoryObject.objects.create(
            name='object2.com', facter_location=facter_location2
        )
        self.obj2_pk = obj2.pk

        self.attr3_pk = obj2.attributes.create(name='attribute1', value='obj2 attribute1 value').pk

        obj2.responsible.add(user1)

        obj2.label.add(label1)

    def test_app_authentication(self):
        authenticated_client = APIClient(HTTP_X_AUTH_APPLICATION='authapp1',
                                         HTTP_X_AUTH_TOKEN='authtoken1')

        response = authenticated_client.get('/api/v2/')

        self.assertEqual(response.status_code, 200)

    def test_user_authentication(self):
        client = APIClient()

        User.objects.create_superuser('admin', 'foo@foo.com', 'admin')
        client.login(username='admin', password='admin')

        response = client.post('/api/v2/objects/', {
            'name': 'object3.com',
        })

        self.assertEqual(response.status_code, 201)

    def test_authenticated_create(self):
        authenticated_client = APIClient(HTTP_X_AUTH_APPLICATION='authapp1',
                                         HTTP_X_AUTH_TOKEN='authtoken1')

        response = authenticated_client.post('/api/v2/objects/', {'name': 'object3.com'})
        self.assertEqual(response.status_code, 201)

        response = authenticated_client.post('/api/v2/objects/', {'name': 'object4'})
        self.assertEqual(response.status_code, 201)

        response = authenticated_client.post('/api/v2/objects/', {'name': 'illegal_hostname.l'})
        self.assertEqual(response.status_code, 400)

    def test_invalid_auth_app(self):
        authenticated_client = APIClient(
            HTTP_X_AUTH_APPLICATION='invalidauthapp',
            HTTP_X_AUTH_TOKEN='authtoken1')

        response = authenticated_client.get('/api/v2/')

        self.assertEqual(response.status_code, 403)

    def test_invalid_auth_token(self):
        authenticated_client = APIClient(HTTP_X_AUTH_APPLICATION='authapp1',
                                         HTTP_X_AUTH_TOKEN='invalidauthtoken')

        response = authenticated_client.get('/api/v2/')

        self.assertEqual(response.status_code, 403)

    def test_inactive_app(self):
        authenticated_client = APIClient(HTTP_X_AUTH_APPLICATION='authapp2',
                                         HTTP_X_AUTH_TOKEN='authtoken2')

        response = authenticated_client.get('/api/v2/')

        self.assertEqual(response.status_code, 200)

        Application.objects.filter(name='authapp2').update(active=False)

        response = authenticated_client.get('/api/v2/')

        self.assertEqual(response.status_code, 403)

    def assertResponseJSONContains(self, response, status_code, expected):
        if response.status_code != status_code:
            message = '{} != {}\n{}'.format(response.status_code, status_code,
                                            response)

            self.fail(message)

        response_json = response.rendered_content
        actual = json.loads(response_json)

        if isinstance(expected, list):
            self.assertIsInstance(actual, list)
            expected_items = expected
            actual_items = actual
        else:
            expected_items = [expected]
            actual_items = [actual]

        # To account for additional fields, which may appear in responses,
        # delete fields which aren't present in expected result before
        # comparison.
        # Can't use assertDictContainsSubset here, because we can have several
        # items here and I want to have diff of full response on test failure.
        for actual_item, expected_item in zip(actual_items, expected_items):
            if isinstance(expected_item, dict) and isinstance(actual_item,
                                                              dict):
                extra_keys = set(actual_item) - set(expected_item)

                for key in extra_keys:
                    del actual_item[key]

        self.assertEqual(actual, expected)

    def test_get_objects(self):
        response = self.client.get('/api/v2/objects/?ordering=id')

        self.assertResponseJSONContains(response, 200, [
            {
                'id': self.obj1_pk,
                'name': 'object1.com',
                'location': {
                    'name': 'location1_name',
                    'fact': 'facter_location1_name'
                },
                'type': {'name': 'type1'},
                'labels': [],
                'attributes': {
                    'attribute1': 'obj1 attribute1 value',
                    'attribute2': 'obj1 attribute2 value',
                },
                'responsible': []
            },
            {
                'id': self.obj2_pk,
                'name': 'object2.com',
                'location': {
                    'name': 'location1_name',
                    'fact': 'facter_location2_name'
                },
                'type': None,
                'labels': ['label1'],
                'attributes': {
                    'attribute1': 'obj2 attribute1 value'
                },
                'responsible': ['user1']
            },
        ])

    def test_create_object(self):
        response = self.client.post('/api/v2/objects/', {
            'name': 'object3.com',
        })

        self.assertResponseJSONContains(response, 201, {
            'name': 'object3.com',
            'location': None,
            'type': None,
            'labels': [],
            'attributes': {},
            'responsible': []
        })

    def test_update_object(self):
        response = self.client.patch('/api/v2/objects/{}/'.format(self.obj1_pk), {
            'name': 'modified-object1.com',
        })

        self.assertEqual(response.status_code, 200)
        self.assertEqual(InventoryObject.objects.get(id=self.obj1_pk).name, 'modified-object1.com')

    def test_create_object_with_attributes(self):
        object_response = self.client.post('/api/v2/objects/', {
            'name': 'object3.com',
            'attributes': {
                'attribute1': 'obj3 attribute1 value',
            }
        }, format='json')

        object_id = object_response.data['id']

        get_response = self.client.get('/api/v2/objects/{}/'.format(object_id))

        self.assertResponseJSONContains(get_response, 200, {
            'id': object_id,
            'name': 'object3.com',
            'location': None,
            'type': None,
            'labels': [],
            'attributes': {
                'attribute1': 'obj3 attribute1 value',
            },
            'responsible': []
        })

    def test_create_user_with_defaults(self):
        create_response = self.client.post('/api/v2/users/',
                                           {'username': 'user2'})

        self.assertResponseJSONContains(create_response, 201, {
            'username': 'user2',
            'url': 'http://testserver/api/v2/users/2/',
            'email': '',
            'groups': []
        })

        response = self.client.get('/api/v2/users/?ordering=id')

        self.assertResponseJSONContains(response, 200, [
            {'id': self.user_pk, 'username': 'user1'},
            {
                'username': 'user2',
                'url': 'http://testserver/api/v2/users/2/',
                'email': '',
            },
        ])

    def test_create_user(self):
        create_response = self.client.post('/api/v2/users/', {
            'username': 'user3',
            'email': 'user3@mail.nonexistent',
        })

        self.assertResponseJSONContains(create_response, 201, {
            'username': 'user3',
            'url': 'http://testserver/api/v2/users/2/',
            'email': 'user3@mail.nonexistent',
        })

        response = self.client.get('/api/v2/users/?ordering=id')

        self.assertResponseJSONContains(response, 200, [
            {'id': self.user_pk, 'username': 'user1'},
            {
                'username': 'user3',
                'url': 'http://testserver/api/v2/users/2/',
                'email': 'user3@mail.nonexistent',
            },
        ])

    def test_objects_name_filter(self):
        response = self.client.get('/api/v2/objects/?name=object2.com')

        self.assertResponseJSONContains(response, 200, [
            {'id': self.obj2_pk},
        ])

    def test_objects_location_fact_filter(self):
        response = self.client.get(
            '/api/v2/objects/?location__fact=facter_location1_name')

        self.assertResponseJSONContains(response, 200, [
            {'id': self.loc_pk},
        ])

    def test_object_type_name_filter(self):
        response = self.client.get(
            '/api/v2/objects/?type__name=type1')

        self.assertResponseJSONContains(response, 200, [
            {'id': self.type_pk},
        ])

    def test_get_users(self):
        response = self.client.get('/api/v2/users/')

        self.assertResponseJSONContains(response, 200, [
            {
                'url': 'http://testserver/api/v2/users/1/',
                'username': 'user1',
                'email': 'user1@mail',
                'groups': [
                ],
            },
        ])

    def test_get_ipv4networks(self):
        response = self.client.get('/api/v2/ipv4networks/')

        self.assertResponseJSONContains(response, 200, [
            {
                'id': self.vlan_pk,
                'address': '10.0.0.1',
                'netmask': '255.0.0.0',
                'gateway': '10.0.0.254',
            }
        ])

    def test_patch_object_labels_format_json(self):
        # Due to bug in implementation, request only works with format='json'
        patch_response = self.client.patch('/api/v2/objects/2/', {
            'labels': ['newlabel1', 'newlabel2', 'newlabel3']
        }, format='json')

        self.assertEqual(patch_response.status_code, 200)

        response = self.client.get('/api/v2/objects/{}/'.format(self.obj2_pk))

        self.assertResponseJSONContains(response, 200, {
            'id': self.obj2_pk,
            'name': 'object2.com',
            'labels': ['newlabel1', 'newlabel2', 'newlabel3'],
        })

    def test_patch_object_responsible_format_json(self):
        # Due to bug in implementation, request only works with format='json'
        User.objects.create(username='user2', email='user2@mail')
        User.objects.create(username='user3', email='user3@mail')

        patch_response = self.client.patch('/api/v2/objects/{}/'.format(self.obj2_pk), {
            'responsible': ['user2', 'user3']
        }, format='json')

        self.assertEqual(patch_response.status_code, 200)

        get_response = self.client.get('/api/v2/objects/{}/'.format(self.obj2_pk))

        self.assertResponseJSONContains(get_response, 200, {
            'name': 'object2.com',
            'responsible': ['user2', 'user3']
        })

    def test_patch_object_location(self):
        response = self.client.patch('/api/v2/objects/{}/'.format(self.obj2_pk), {
            'location': {'fact': 'newfact'}
        }, format='json')

        self.assertEqual(response.status_code, 200)

        obj = InventoryObject.objects.get(id=self.obj2_pk)

        self.assertEqual(obj.facter_location.fact, 'newfact')

    def test_patch_object_location_to_existing_fact(self):
        obj = InventoryObject.objects.get(id=self.obj2_pk)

        self.assertEqual(obj.facter_location.fact, 'facter_location2_name')

        response = self.client.patch('/api/v2/objects/{}/'.format(self.obj2_pk), {
            'location': {'fact': 'facter_location1_name'}
        }, format='json')

        self.assertEqual(response.status_code, 200)

        obj = InventoryObject.objects.get(id=self.obj2_pk)
        self.assertEqual(obj.facter_location.fact, 'facter_location1_name')

    def test_patch_object_location_validation(self):
        response = self.client.patch('/api/v2/objects/{}/'.format(self.obj2_pk), {
            'location': {}
        }, format='json')
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.data, {"location": ["Only fact can be set via API"]})

        response = self.client.patch('/api/v2/objects/{}/'.format(self.obj2_pk), {
            'location': {'name': 'testname', 'fact': 'testfact'}
        }, format='json')
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.data, {"location": ["Only fact can be set via API"]})

    def test_patch_object_type(self):
        response = self.client.patch('/api/v2/objects/{}/'.format(self.obj2_pk), {
            'type': {'name': 'newtype'},
        }, format='json')
        self.assertEqual(response.status_code, 200)

        obj = InventoryObject.objects.get(id=self.obj2_pk)
        self.assertEqual(obj.type.name, 'newtype')


@unittest.skip("All API tests temporarily disabled")
class UnprivilegedUserTestCase(APITestCase):
    maxDiff = None

    # Primary keys for test fixtures (used in URLs and some asserts)
    obj1_pk = None
    attr1_pk = None
    attr2_pk = None
    label_pk = None

    object_permissions_mock = patch(
        'inventory.models.InventoryObject.get_permissions',
        return_value=Permissions(view=('id', 'name', 'label', 'attributes'), edit=('name', 'label'))
    )
    object_permissions_mock_no_edit_labels = patch(
        'inventory.models.InventoryObject.get_permissions',
        return_value=Permissions(view=('id', 'name', 'label', 'attributes'), edit=('name',))
    )
    attribute_permissions_mock = patch(
        'inventory.models.InventoryObjectAttribute.get_permissions',
        return_value=Permissions(edit=('value',)), create=True)

    def setUp(self):
        user1 = User.objects.create_user(username='user1', password='password1',
                                         email='user1@mail')

        self.client.force_authenticate(user=user1)

        obj1 = InventoryObject.objects.create(
            name='object1.com',
        )
        self.obj1_pk = obj1.pk

        self.attr1_pk = \
            obj1.attributes.create(name='attribute1', value='obj1 attribute1 value').pk
        self.attr2_pk = \
            obj1.attributes.create(name='attribute2', value='obj1 attribute2 value').pk

        label1 = InventoryObjectLabel.objects.create(name='label1')
        self.label_pk = label1.pk

        obj1.label.add(label1)

    def test_hidden_values(self):
        with self.object_permissions_mock:
            response = self.client.get('/api/v2/objects/{}/'.format(self.obj1_pk))

            self.assertEqual(response.status_code, 200)

            self.assertEqual(response.data, {
                "id": self.obj1_pk,
                "name": "object1.com",
                "location": "__access_denied__",
                "type": "__access_denied__",
                "labels": ["label1"],
                "attributes": {
                    "attribute1": "obj1 attribute1 value",
                    "attribute2": "obj1 attribute2 value",
                },
                "responsible": "__access_denied__",
            })

    def test_related_field_partial_update(self):
        with self.object_permissions_mock:
            response = self.client.patch('/api/v2/objects/{}/'.format(self.obj1_pk), {
                'name': 'newname.com',
                'labels': ['newlabel1', 'newlabel2'],
            }, format='json')

            self.assertEqual(response.status_code, 200)

            obj1 = InventoryObject.objects.get(id=self.obj1_pk)

            self.assertEqual(obj1.name, 'newname.com')

            self.assertItemsEqual(obj1.label.values_list('name', flat=True),
                                  ['newlabel1', 'newlabel2'])

    def test_related_field_partial_update_no_permission(self):
        with self.object_permissions_mock_no_edit_labels:
            response = self.client.patch('/api/v2/objects/{}/'.format(self.obj1_pk), {
                'name': 'newname.com',
                'labels': ['newlabel1', 'newlabel2'],
            }, format='json')

            self.assertEqual(response.status_code, 200)

            obj1 = InventoryObject.objects.get(id=self.obj1_pk)

            self.assertEqual(obj1.name, 'newname.com')

            self.assertItemsEqual(obj1.label.values_list('name', flat=True),
                                  ['label1'])
