import csv
from collections import OrderedDict

from django.contrib.auth.decorators import login_required
from django.contrib.auth.models import User
from django.shortcuts import render
from django.shortcuts import HttpResponse

from inventory.models import (
    InventoryObject,
    InventoryObjectAttribute,
    InventoryObjectLabel,
)


@login_required
def search(request):
    query = request.GET.get('query').strip()
    user = request.user

    if query.startswith('^') and query.endswith('$'):
        search_type = 'iexact'
        query = query[1:-1]
    elif query.startswith('^'):
        search_type = 'istartswith'
        query = query[1:]
    elif query.endswith('$'):
        search_type = 'iendswith'
        query = query[:-1]
    else:
        search_type = 'icontains'

    attributes = _search_for_inventory_objects(search_type, query, user)
    object_attrs = _search_for_inventory_object_attributes(search_type, query)

    for name in object_attrs:
        if name in attributes:
            attributes[name]['results'] += object_attrs[name]['results']
        else:
            attributes[name] = object_attrs[name]

    return render(request, 'inventory/search.html', {
        'attributes': attributes,
        'query': query,
    })


@login_required
def export_object_inventory(request):
    label = request.GET.get('label')
    responsible = request.GET.get('responsible')

    filters = {}
    if label is not None:
        filters['label'] = InventoryObjectLabel.objects.get(name=label)

    if responsible is not None:
        filters['responsible'] = User.objects.get(username=responsible)

    object_list = InventoryObject.objects.filter(
        name__isnull=False, **filters
    ).select_related(
        'facter_location__location', 'type', 'assigned_user', 'assigned_group',
    ).prefetch_related(
        'attributes', 'label', 'responsible', 'labs', 'jenkins_nodes__instance',
    )

    columns = OrderedDict([
        ('name', 'Name'),
        ('location_name', 'Location'),
        ('location_fact', 'Location Fact'),
        ('hardware', 'Hardware'),
        ('type', 'Type'),
        ('label_string', 'Labels'),
        ('jira_issue', 'Jira Issue'),
        ('dcim_jira_issue', 'DCIM Jira Issue'),
        ('responsible_string', 'Responsible Engineers'),
        ('assigned_user_name', 'Assigned User'),
        ('assigned_group_name', 'Assigned Group'),
        ('used_in_string', 'Used In'),
    ])

    response = HttpResponse(
        content_type='text/csv',
        reason='Ok',
    )

    writer = csv.writer(response)
    writer.writerow(columns.values())
    for i in object_list:
        writer.writerow([getattr(i, c) for c in columns.iterkeys()])

    return response


def _search_for_inventory_objects(search_type, query, user):
    fields_mapping = InventoryObject.SEARCHABLE_FIELDS_MAPPING.items()
    attributes = {}

    for display_name, attr_name in fields_mapping:
        search_query = {'{}__{}'.format(attr_name, search_type): query}
        inventory_objs = [
            obj for obj in InventoryObject.objects.filter(**search_query)
            if InventoryObject.get_permissions(user, obj).get_field_permissions(
                attr_name.split('__')[0]).view
        ]

        for inventory_object in inventory_objs:
            value = inventory_object

            for child in attr_name.split('__'):
                value = getattr(value, child)

            data = {
                'name': display_name,
                'value': value,
            }
            name = inventory_object.display_name
            if name in attributes:
                attributes[name]['results'].append(data)
            else:
                attributes[name] = {
                    'object': inventory_object,
                    'results': [data],
                }

    return attributes


def _search_for_inventory_object_attributes(search_type, query):
    search_query = {'value__{}'.format(search_type): query}
    attributes = {}

    inventory_attributes_objs = InventoryObjectAttribute.objects.\
        filter(**search_query).select_related('object')

    for attribute in inventory_attributes_objs:
        _obj = attribute.object

        attributes[_obj.display_name] = {
            'object': _obj,
            'results': [{
                'name': attribute.name,
                'value': attribute.awesome_value,
            }],
        }

    return attributes
