from django.db import models

from racks.models import RacksModel, RacksTimestampedModel
from inventory.fields import NullableCharField, EncryptedCharField
from security import permissions


class JenkinsInstance(RacksTimestampedModel):

    name = models.CharField(max_length=32, unique=True)
    url = models.CharField(max_length=255, unique=True, verbose_name='URL')
    username = NullableCharField(max_length=63, default=None, null=True, blank=True)
    api_key = EncryptedCharField(max_length=255, default=None, null=True, blank=True,
                                 verbose_name='API key')

    last_successful_sync = models.DateTimeField(default=None, null=True)
    last_failed_sync = models.DateTimeField(default=None, null=True)

    master_node = models.ForeignKey(
        'JenkinsNode',
        null=True,
        related_name='+',
        on_delete=models.SET_NULL,
    )

    def __repr__(self):
        return '<{}: {} @ {}>'.format(type(self).__name__, self.name, self.url)

    @property
    def failed_sync(self):
        if self.last_successful_sync is None:
            if self.last_failed_sync is None:
                return False
            else:
                return True
        else:
            if self.last_failed_sync is None:
                return False
            else:
                return self.last_successful_sync < self.last_failed_sync

    PUBLIC_FIELDS = (
        'id',
        'name',
        'url',
        'last_successful_sync',
        'last_failed_sync',
        'master_node',
        'nodes',
    )

    @classmethod
    def get_permissions(cls, user, obj):
        return permissions.Permissions(view=cls.PUBLIC_FIELDS)


class JenkinsLabel(RacksModel):

    name = models.CharField(max_length=63, unique=True)

    def __repr__(self):
        return '<{}: {}>'.format(type(self).__name__, self.name)


class JenkinsNode(RacksTimestampedModel):

    instance = models.ForeignKey(
        JenkinsInstance,
        null=True,
        blank=True,
        related_name='nodes',
        on_delete=models.SET_NULL,
    )
    inventory_object = models.ForeignKey(
        'inventory.InventoryObject',
        null=True,
        blank=True,
        related_name='jenkins_nodes',
        on_delete=models.SET_NULL,
    )
    labels = models.ManyToManyField(
        JenkinsLabel,
        related_name='nodes',
    )

    name = models.CharField(max_length=127, db_index=True)

    hostname = models.CharField(max_length=127, verbose_name='Hostname', null=True, db_index=True)
    address = models.CharField(max_length=15, verbose_name='IPv4/6 Address', null=True)

    offline = models.BooleanField(default=False)
    temporarily_offline = models.BooleanField(default=False)
    offline_reason = NullableCharField(max_length=127, null=True, default=None)

    architecture = NullableCharField(max_length=63, null=True, default=None)
    executors = models.IntegerField(default=0)

    available_memory = models.BigIntegerField(default=0)
    total_memory = models.BigIntegerField(default=0)
    available_swap = models.BigIntegerField(default=0)
    total_swap = models.BigIntegerField(default=0)
    disk_space = models.BigIntegerField(default=0)
    temp_disk_space = models.BigIntegerField(default=0)

    last_job = models.DateTimeField(null=True, default=None)

    disabled = models.BooleanField(null=False, default=False)

    def __repr__(self):
        return '<{}: {} ({}{})>'.format(
            type(self).__name__,
            self.name,
            self.formatted_status,
            ', disabled' if self.disabled else '',
        )

    @property
    def formatted_address(self):
        if self.hostname:
            return self.hostname
        elif self.address:
            return self.address
        else:
            return '(no host/IP available)'

    @property
    def formatted_status(self):
        if self.offline:
            return 'offline'
        elif self.temporarily_offline:
            return 'temp offline'
        return 'online'

    @property
    def all_labels(self, separator=u' '):
        return separator.join([label.name for label in self.labels.all().order_by('name')])


class JenkinsErrorLog(RacksModel):

    created = models.DateTimeField(auto_now_add=True)

    instance = models.ForeignKey(
        JenkinsInstance,
        null=True,
        blank=True,
        related_name='error_logs',
        on_delete=models.SET_NULL,
    )

    instance_name = models.CharField(max_length=32)
    instance_url = models.CharField(max_length=127)
    message = models.CharField(max_length=255)

    def __repr__(self):
        return '<%s: [%s] %s>' % (type(self).__name__, self.instance_name, self.message)
