from django.conf import settings
from django.core.paginator import Paginator

from racks import crud
from jenkins_ci.forms import JenkinsInstanceForm
from jenkins_ci.models import JenkinsInstance, JenkinsNode
from security.permissions import get_permissions


class JenkinsInstanceListView(crud.ListView):
    template_name = 'jenkins_ci/instances/change_list.html'
    display_actions = not settings.JENKINS_DISABLE_UI_FORMS

    def get_queryset(self):
        return JenkinsInstance.objects.all()


class JenkinsInstanceDetailView(crud.DetailView):
    template_name = 'jenkins_ci/instances/detail.html'
    paginate_by = 20
    display_actions = not settings.JENKINS_DISABLE_UI_FORMS

    def get_context_data(self, **kwargs):
        context = super(JenkinsInstanceDetailView, self).get_context_data(**kwargs)

        # `tab` parameter is normally used to tell what tab should be displayed.
        # Normally defaults to `object` (i.e. first tab) except when `page` parameter is present,
        # which means we should be showing `nodes` (i.e. second tab, "Slaves").
        tab = self.request.GET.get('tab', 'nodes' if self.request.GET.get('page') else 'object')

        queryset_nodes = self.object.nodes.order_by('name')
        nodes_paginator = Paginator(queryset_nodes, self.paginate_by)
        nodes_page = nodes_paginator.page(self.request.GET.get('page', 1))

        get_parameters = self.request.GET.copy()
        get_parameters.pop('page', None)

        queryset_logs = self.object.error_logs.order_by('-created')[0:50]

        context.update({
            'tab': tab,
            'is_paginated': True,
            'nodes_list': nodes_page.object_list,
            'nodes_paginator': nodes_paginator,
            'nodes_page': nodes_page,
            'get_parameters': get_parameters,
            'logs_list': queryset_logs,
            'can_edit_nodes': get_permissions(self.request, JenkinsNode).create
        })
        return context


class JenkinsInstanceDeleteView(crud.DeleteView):

    def delete_object(self, instance):
        if instance.master_node:
            instance.master_node.delete()
        instance.delete()


class JenkinsInstanceViewSet(crud.ViewSet):
    model = JenkinsInstance
    list_view = JenkinsInstanceListView
    detail_view = JenkinsInstanceDetailView

    if settings.JENKINS_DISABLE_UI_FORMS:
        create_view = None
        update_view = None
        change_form_class = None
        delete_view = None
    else:
        change_form_class = JenkinsInstanceForm
        delete_view = JenkinsInstanceDeleteView

    readonly_fields = ('failed_sync', 'master_node')


urlpatterns = JenkinsInstanceViewSet('jenkins_ci.instances').urlpatterns
