from django import forms
from django.conf.urls import url
from django.contrib.auth.decorators import login_required
from django.core.exceptions import ImproperlyConfigured
from django.core.paginator import Paginator
from django.core.urlresolvers import reverse
from django.forms.forms import BoundField
from django.forms.models import modelform_factory
from django.forms import widgets
from django.http import HttpResponseForbidden, HttpResponseRedirect
from django.shortcuts import get_object_or_404
from django.views.generic.base import TemplateView
from django.utils.decorators import method_decorator, classonlymethod

from racks.widgets import FilteredSelectMultiple
from security.permissions import ProtectedObject, get_permissions


class CRUDView(TemplateView):
    view_set = None
    model = None
    template = None

    @property
    def url_names(self):
        return self.view_set.url_names

    def get_model_verbose_name(self):
        return self.model._meta.verbose_name

    @method_decorator(login_required)
    def dispatch(self, request, *args, **kwargs):
        self.permissions = get_permissions(request, self.model)

        if not self.view_allowed(self.permissions):
            return HttpResponseForbidden()

        return super(CRUDView, self).dispatch(request, *args, **kwargs)

    @classonlymethod
    def as_view(self, view_set):
        return super(CRUDView, self).as_view(view_set=view_set, model=view_set.model)

    def get_template_names(self):
        if self.template_name:
            return [self.template_name]

        path = self.view_set.prefix.replace('.', '/')
        if not self.template:
            raise ImproperlyConfigured("'template_name' or 'template' variable should be defined")
        template = self.template

        return [
            '{}/{}.html'.format(path, template),
            'crud/{}.html'.format(template),
        ]

    def view_allowed(self, permissions):
        return True

    def get_queryset(self):
        return self.model.objects.all()


class FieldForDisplay(object):
    def __init__(self, label):
        self.label = label


class ReadOnlyBoundField(BoundField):
    def as_widget(self, widget=None, attrs=None, only_initial=False):
        attrs = attrs or {}
        attrs['readonly'] = 'readonly'

        widget = widget or self.field.widget

        # Disable opening of Select fields. We're checking for type, because our
        # FilteredSelectMultiple widget is subclass of Select, but correctly handles readonly
        # attribute.
        if type(widget) is widgets.Select:
            attrs['disabled'] = True

        return super(ReadOnlyBoundField, self).as_widget(widget, attrs, only_initial)


class CRUDFormView(CRUDView):

    def get_initial(self, obj):
        initial = {}

        for field_name in self.view_set.related_foreign_key_fields:
            if obj:
                initial[field_name] = getattr(obj, field_name).all()

        return initial

    def get_form_class(self):
        if self.view_set.change_form_class:
            return self.view_set.change_form_class

        widgets = {
            field: FilteredSelectMultiple
            for field in self.view_set.filtered_select_fields
        }

        if self.view_set.widgets:
            widgets.update(self.view_set.widgets)

        return modelform_factory(self.model, fields=self.view_set.fields, widgets=widgets,
                                 formfield_callback=self.formfield_callback)

    def get_fields_for_form_display(self, form):
        editable_fields = {}
        fields_for_display = []

        for field_name, field in form.fields.items():
            bound_field = form[field_name]
            field_for_display = FieldForDisplay(bound_field.label)
            field_permissions = self.permissions.get_field_permissions(field_name)

            editable = (field_permissions.edit and field_name not in self.view_set.readonly_fields)
            readable = field_permissions.view

            field_for_display.form_field = bound_field

            if editable:
                editable_fields[field_name] = field
            elif readable:
                field_for_display.form_field = ReadOnlyBoundField(form, field, field_name)
            else:
                field_for_display.no_permission = True

            fields_for_display.append(field_for_display)

        return fields_for_display, editable_fields

    def get_form(self, obj=None, data=None):
        form_class = self.get_form_class()
        form = form_class(data=data, instance=obj, initial=self.get_initial(obj))

        for field_name in self.view_set.related_foreign_key_fields:
            field = getattr(self.model, field_name)

            qs = field.related.model.objects.all()

            form.fields[field_name] = forms.ModelMultipleChoiceField(
                qs, required=False, widget=FilteredSelectMultiple
            )

        form.fields_for_display, form.fields = self.get_fields_for_form_display(form)

        # If no fields are editable, disable save
        form.allow_save = bool(form.fields)

        return form

    def form_valid(self, form):
        response = super(CRUDFormView, self).form_valid(form)

        for field_name in self.view_set.related_foreign_key_fields:
            value = form.cleaned_data[field_name]

            setattr(self.object, field_name, value)

        return response

    def formfield_callback(self, f, **kwargs):
        field_class = self.view_set.field_classes.get(f.name)

        if field_class:
            return field_class(**kwargs)

        return f.formfield(**kwargs)

    def get_success_url(self, obj):
        try:
            return obj.get_absolute_url()
        except AttributeError:
            return reverse(self.url_names['list'])

    def post(self, request, **kwargs):
        context = self.get_context_data(**kwargs)
        form = context['form']

        if not form.allow_save:
            return HttpResponseForbidden()

        if form.is_valid():
            obj = form.save()
            return HttpResponseRedirect(self.get_success_url(obj))

        return self.render_to_response(context)


class ListView(CRUDView):
    paginate_by = 20
    template = 'change_list'
    display_actions = True  # Display Actions column or not

    def get_parameters(self):
        get_parameters = self.request.GET.copy()
        get_parameters.pop('page', None)
        return get_parameters

    def get_context_data(self, **kwargs):
        context = super(ListView, self).get_context_data(**kwargs)
        try:
            per_page = int(self.request.GET.get('per_page', self.paginate_by))
        except ValueError:
            per_page = self.paginate_by

        try:
            page_no = int(self.request.GET.get('page', 1))
        except ValueError:
            page_no = 1

        queryset = self.get_queryset()
        paginator = Paginator(queryset, per_page)
        page = paginator.page(page_no)

        context.update({
            'page_obj': page,
            'paginator': paginator,
            'object_list': page.object_list,
            'permissions': self.permissions,
        })
        return context


class DetailView(CRUDView):
    object = None
    permissions = None
    template = 'detail'

    def get_object(self, **kwargs):
        return get_object_or_404(self.get_queryset(), pk=kwargs['pk'])

    def get_context_data(self, **kwargs):
        context = super(DetailView, self).get_context_data(**kwargs)

        self.object = self.get_object(**kwargs)

        self.permissions = get_permissions(self.request, self.model, self.object)
        self.protected_object = ProtectedObject(self.object, self.permissions)
        context['object'] = self.protected_object
        return context


class UpdateView(CRUDFormView):
    template_name = 'crud/update_form.html'
    object = None

    def get_context_data(self, **kwargs):
        context = super(UpdateView, self).get_context_data(**kwargs)

        self.object = get_object_or_404(self.get_queryset(), pk=kwargs['pk'])
        self.permissions = get_permissions(self.request, self.model, self.object)
        if self.request.method == 'POST':
            context['form'] = self.get_form(obj=self.object, data=self.request.POST)
        else:
            context['form'] = self.get_form(obj=self.object)
        context['object'] = ProtectedObject(self.object, self.permissions),
        return context


class CreateView(CRUDFormView):
    template_name = 'crud/create_form.html'

    def view_allowed(self, permissions):
        return permissions.create

    def get_context_data(self, **kwargs):
        context = super(CreateView, self).get_context_data(**kwargs)
        if self.request.method == 'POST':
            context['form'] = self.get_form(data=self.request.POST)
        else:
            context['form'] = self.get_form()
        return context


class DeleteView(CRUDView):
    template_name = 'crud/confirm_delete.html'

    def view_allowed(self, permissions):
        return permissions.delete

    def get_success_url(self):
        return self.request.GET.get('next') or reverse(self.router.list_view.url_name)

    def delete_object(self, obj):
        """Actually performs object deletion. To be overridden in child classes if necessary."""
        obj.delete()

    def post(self, request, pk):
        obj = get_object_or_404(self.get_queryset(), pk=pk)
        self.delete_object(obj)
        return HttpResponseRedirect(self.get_success_url())


class ViewSet(object):
    PATTERNS = (
        (r'^$', 'list'),
        (r'^(?P<pk>[0-9]+)$', 'detail'),
        (r'^new$', 'create'),
        (r'^(?P<pk>[0-9]+)/edit$', 'update'),
        (r'^(?P<pk>[0-9]+)/delete$', 'delete'),
    )

    list_view = ListView
    detail_view = None
    create_view = CreateView
    update_view = UpdateView
    delete_view = DeleteView

    change_form_class = None
    filtered_select_fields = ()
    readonly_fields = ()
    related_foreign_key_fields = ()

    fields = None
    field_classes = {}
    widgets = None

    def __init__(self, prefix):
        self.prefix = prefix
        self.url_names = {}
        self.urlpatterns = []

        for pattern, name in self.PATTERNS:
            view_attr = name + '_view'
            view_class = getattr(self, view_attr)

            if view_class:
                url_name = '%s.%s' % (prefix, name)
                self.url_names[name] = url_name
                self.urlpatterns.append(url(pattern, view_class.as_view(self), name=url_name))

            self.url_names['object'] = self.url_names.get('detail') or self.url_names.get('update')
