import base64

from django.conf import settings
from django.utils.six import binary_type, text_type

from cryptography.fernet import Fernet, InvalidToken
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC


class EncryptionError(ValueError):
    pass


class EnigmaMachine(object):

    _fernet = None

    def __init__(self, password, salt=''):
        kdf = PBKDF2HMAC(
            algorithm=hashes.SHA256(),
            length=32,
            salt=salt,
            iterations=100000,
            backend=default_backend(),
        )
        key = base64.urlsafe_b64encode(kdf.derive(password))
        self._fernet = Fernet(key)

    def encrypt(self, value):
        if not value:
            return None
        # Enforce binary data
        value = value.encode('utf-8') if isinstance(value, text_type) else binary_type(value)
        return self._fernet.encrypt(value)

    def decrypt(self, value):
        if not value:
            return value
        # Enforce binary data
        value = value.encode('utf-8') if isinstance(value, text_type) else binary_type(value)
        try:
            return self._fernet.decrypt(value)
        except InvalidToken:
            raise EncryptionError

    def decrypt_or_return(self, encrypted):
        try:
            return self.decrypt(encrypted)
        except EncryptionError:
            return encrypted


enigma = EnigmaMachine(settings.SECRET_KEY)
