from django.core.urlresolvers import reverse
from django.contrib.auth.models import User


class Section(object):
    def __init__(self, label, subsections):
        self.label = label
        self.subsections = subsections

    def __iter__(self):
        return iter(self.subsections)

    @property
    def url(self):
        return self.subsections[0].url


class Subsection(object):
    def __init__(self, label, main_url_name, url_name_prefix=None, **kwargs):
        self.label = label
        self.main_url_name = main_url_name
        self.url_name_prefix = url_name_prefix
        self.kwargs = kwargs

    @property
    def url(self):
        return reverse(self.main_url_name, kwargs=self.kwargs)


class Navigation(object):
    def __init__(self, sections, active_section, active_subsection):
        self.sections = sections
        self.active_section = active_section
        self.active_subsection = active_subsection


def sections(request, view_func, view_args, view_kwargs):
    sections_list = [
        Section('Inventory', [
            Subsection('Objects', 'inventory.objects.list', 'inventory.objects.'),
            Subsection('Labs', 'inventory.labs.list', 'inventory.labs.'),
            Subsection('Labels', 'inventory.labels.list', 'inventory.labels.'),
            Subsection('Locations', 'inventory.locations.list', 'inventory.locations.'),
            Subsection('IPv4 Networks', 'inventory.ipv4networks.list', 'inventory.ipv4networks.'),
            Subsection('VLANs', 'inventory.vlans.list', 'inventory.vlans.'),
            Subsection('Switches', 'inventory.switches.list', 'inventory.switches.'),
        ]),
        Section('Jenkins', [
            Subsection('Jenkins Instances', 'jenkins_ci.instances.list', 'jenkins_ci.instances.'),
            Subsection('Jenkins Slaves', 'jenkins_ci.nodes.list', 'jenkins_ci.nodes.'),
        ]),
        Section('Reports', [
            Subsection('Hosts by location', 'reports.views.basic'),
            Subsection('Jenkins slaves', 'reports.views.jenkins'),
        ]),
    ]
    user_sections = []
    if 'username' in view_kwargs:
        try:
            user = User.objects.get_by_natural_key(view_kwargs['username'])
        except User.DoesNotExist:
            pass
        else:
            if user != request.user:
                user_sections = [
                    Subsection('Profile', 'account.profile', username=user.username),
                    Subsection('Assigned', 'account.assigned', username=user.username),
                    Subsection('Responsible', 'account.responsible', username=user.username),
                ]
    sections_list.append(Section('Users', [
        Subsection('List', 'account.list')
    ] + user_sections))
    sections_list.append(Section('Account', [
        Subsection('Profile', 'account.profile', username=request.user.username),
        Subsection('Assigned', 'account.assigned', username=request.user.username),
        Subsection('Responsible', 'account.responsible', username=request.user.username),
        Subsection('Applications', 'security.applications.list', 'security.applications.'),
    ]))
    return sections_list


class NavigationMiddleware:
    def get_active(self, request, sections_list):
        for section in sections_list:
            for subsection in section.subsections:
                url_name = request.resolver_match.url_name
                # url_name can be None in 1.8, fixed the condition appropriately
                if url_name is None:
                    continue
                if (url_name == subsection.main_url_name) or subsection.url_name_prefix \
                        and url_name.startswith(subsection.url_name_prefix):
                    return section, subsection
        return None, None

    def process_view(self, request, view_func, view_args, view_kwargs):
        sections_list = sections(request, view_func, view_args, view_kwargs)
        section, subsection = self.get_active(request, sections_list)
        request.navigation = Navigation(sections_list, section, subsection)
